import copy
import logging
import os

from cryptography.fernet import Fernet

from ..persistence import read_json_file, write_json_file


logger = logging.getLogger(__name__)


class GenericDataManager:
    """
    A GenericDataManager manages json-data in self.data in a json file. It's meant to
    be derived from. See OtpManager.

    Supports encrypting individual attributes rest. Expects a Fernet key that is unique
    to the user to encrypt/decrypt.
    """

    name_of_json_file = "some_data.json"
    encrypted_fields = []

    # Track any changes to our encryption implementation so we can migrate up any older
    #   data as we go.
    encrypted_storage_version = 1

    @classmethod
    def initial_data(cls):
        return {}

    def __init__(self, data_folder, encryption_key=None):
        # encryption_key indicates that the encrypted_fields (if any) need to be
        #   encrypted at rest.
        self.data_folder = data_folder
        self.encryption_key = encryption_key
        self.load()

    @property
    def data_file(self):
        return os.path.join(self.data_folder, self.__class__.name_of_json_file)

    def load(self):
        # if whatever-the-name.json file exists - load from it
        if os.path.isfile(self.data_file):
            logger.debug(f"Loading existing file {self.data_file}")
            self.data = read_json_file(self.data_file)
        # otherwise - create one and assign unique id
        else:
            logger.debug(f"{self.data_file} doesn't exist. Creating ...")
            self.data = self.__class__.initial_data()
            self._save()

        if self.encryption_key:
            # Decrypt the encrypted fields and store in memory as plaintext
            fernet = Fernet(self.encryption_key)

            if (
                "encrypted_storage_version" in self.data
                and self.encrypted_storage_version
                != self.data["encrypted_storage_version"]
            ):
                raise Exception(
                    "Upgrading a previous encryption version is not yet implemented"
                )

            for attr in self.encrypted_fields:
                if attr in self.data and self.data[attr] is not None:
                    self.data[attr] = fernet.decrypt(self.data[attr].encode()).decode()

    def _save(self):
        if self.encryption_key:
            # Preserve the in-memory data but encrypt a copy to write to disk
            fernet = Fernet(self.encryption_key)
            output_dict = copy.deepcopy(self.data)
            for attr in self.encrypted_fields:
                if attr in output_dict and output_dict[attr] is not None:
                    output_dict[attr] = fernet.encrypt(
                        output_dict[attr].encode()
                    ).decode()
        else:
            output_dict = self.data

        if self.encrypted_fields:
            output_dict["encrypted_storage_version"] = self.encrypted_storage_version

        write_json_file(output_dict, self.data_file)



"""
Here Configuration of your Extension takes place
"""


class BaseConfig:
    """This is a extension-based Config which is used as Base"""

    DEVELOPER_JAVASCRIPT_PYTHON_CONSOLE = False


class DevelopmentConfig(BaseConfig):
    DEVELOPER_JAVASCRIPT_PYTHON_CONSOLE = True



import logging
import os
import shutil

from urllib3.exceptions import NewConnectionError
from requests.exceptions import ConnectionError

from cryptoadvance.specter.specter_error import SpecterError
from ...config import BaseConfig
from ..specter_migrator import SpecterMigration
from ...helpers import load_jsons
from ...managers.node_manager import NodeManager
import requests

logger = logging.getLogger(__name__)


class SpecterMigration_0001(SpecterMigration):
    version = "v1.6.1"  # the version this migration has been rolled out
    # irrelevant though because we'll execute this script in any case
    # as we can't have yet a say on when specter has been started first

    def should_execute(self):
        # This Migration cannot rely on the default-mechanism as the migration_framework was not
        # in place when the functionality has been implemented
        return True

    @property
    def description(self) -> str:
        return """Single-Node migration:
            In v1.3.1 Single Node implementation has been implemented
            Later we had multiple nodes. This migrates the single installation to one of many.
            Effectively it will:
            * Check whether an internal node was existing in ~/.specter/.bitcoin
            * Check whether a new internal default node (bitcoin/main) is NOT existing
            * Move the ~/.specter/.bitcoin to ~/.specter/nodes/specter_bitcoin/.bitcoin-main
            * Creates a json-definition in ~/.specter/nodes/specter_bitcoin.json
        """

    def execute(self):
        source_folder = os.path.join(self.data_folder, ".bitcoin")
        if not os.path.isdir(source_folder):
            logger.info(
                "No .bitcoin directory found in {self.data_folder}. Nothing to do"
            )
            return
        if not os.path.isdir(os.path.join(self.data_folder, "bitcoin-binaries")):
            raise SpecterError(
                "Could not proceed with migration as bitcoin-binaries are not existing."
            )
        if not self._check_port_free():
            logger.error(
                "There is already a Node with the default port configured or running. Won't migrate!"
            )
            return
        # The version will be the version shipped with specter
        bitcoin_version = BaseConfig.INTERNAL_BITCOIND_VERSION
        logger.info(f".bitcoin directory detected in {self.data_folder}. Migrating ...")
        recommended_name = self._find_appropriate_name()
        target_folder = os.path.join(self.data_folder, "nodes", recommended_name)
        logger.info(f"Migrating to folder {target_folder}")
        os.makedirs(target_folder)
        logger.info(f"Moving .bitcoin to folder {target_folder}")
        shutil.move(source_folder, os.path.join(target_folder, ".bitcoin-main"))
        if os.path.isdir(os.path.join(source_folder, "bitcoin.conf")):
            logger.info("Removing bitcoin.conf file")
            os.remove(os.path.join(source_folder, "bitcoin.conf"))
        definition_file = os.path.join(target_folder, "specter_bitcoin.json")
        logger.info(
            f"Creating {definition_file}. This will cause some warnings and even errors about not being able to connect to the node which can be ignored."
        )
        nm = NodeManager(
            data_folder=os.path.join(self.data_folder, "nodes"),
            bitcoind_path=os.path.join(
                self.data_folder, "bitcoin-binaries", "bin", "bitcoind"
            ),
            internal_bitcoind_version=bitcoin_version,
        )
        # Should create a json (see fullpath) like the one below:
        node = nm.add_internal_node(recommended_name)

        # {
        #     "name": "Specter Bitcoin",
        #     "alias": "specter_bitcoin",
        #     "autodetect": false,
        #     "datadir": "/home/someuser/.specter/nodes/specter_bitcoin/.bitcoin-main",
        #     "user": "bitcoin",
        #     "password": "3ah0yc-2dDEwUSqHuuZi-w",
        #     "port": 8332,
        #     "host": "localhost",
        #     "protocol": "http",
        #     "external_node": false,
        #     "fullpath": "/home/someuser/.specter/nodes/specter_bitcoin.json",
        #     "bitcoind_path": "/home/someuser/.specter/bitcoin-binaries/bin/bitcoind",
        #     "bitcoind_network": "main",
        #     "version": "0.21.1"
        # }

    def _find_appropriate_name(self):
        if not os.path.isdir(os.path.join(self.data_folder, "nodes")):
            return "specter_bitcoin"
        if not os.path.isdir(
            os.path.join(self.data_folder, "nodes", "specter_bitcoin")
        ):
            return "specter_bitcoin"
        # Hmm, now it gets a bit trieckier
        if not os.path.isdir(
            os.path.join(self.data_folder, "nodes", "specter_migrated")
        ):
            return "specter_migrated"
        # Now it's getting fishy
        raise SpecterError(
            "I found a node called 'specter_migrated'. This migration script should not run twice."
        )

    def _check_port_free(self, port=8332):
        # For external nodes, we assume that there are already running
        try:
            result = requests.get(f"http://localhost:{port}")
            return False
        except (ConnectionRefusedError, ConnectionError, NewConnectionError):
            pass
        # Now let's check internal Nodes
        if os.path.isfile(os.path.join(self.data_folder, "nodes")):
            configs = load_jsons(os.path.join(self.data_folder, "nodes"))
            ports = [node.port for node in configs.keys()]
            if port in ports:
                return False
        return True



import logging
import os
import sys
from importlib import import_module

from flask import current_app as app
from flask.blueprints import Blueprint
from cryptoadvance.specter.wallet import Wallet
from cryptoadvance.specter.util.reflection import get_template_static_folder
from flask_babel import lazy_gettext as _
from typing import List

from .service_encrypted_storage import ServiceEncryptedStorageManager
from .service_annotations_storage import ServiceAnnotationsStorage

from cryptoadvance.specter.addresslist import Address
from cryptoadvance.specter.services import callbacks


logger = logging.getLogger(__name__)


devstatus_alpha = "alpha"
devstatus_beta = "beta"
devstatus_prod = "prod"


class Service:
    """A base class for Services"""

    # These should be overrided in implementation classes
    id = None
    name = None
    icon = None
    logo = None
    desc = None  # TODO: rename to "description" to be explicit
    has_blueprint = True  # the default
    # If the blueprint gets a "/ext" prefix (isolated_client = True), the login cookie won't work for all specter core functionality
    isolated_client = True
    devstatus = devstatus_alpha

    def __init__(self, active, specter):
        if not hasattr(self, "id"):
            raise Exception(f"Service {self.__class__} needs ID")
        if not hasattr(self, "name"):
            raise Exception(f"Service {self.__class__} needs name")
        self.active = active
        self.specter = specter

    def callback(self, callback_id, *argv, **kwargv):
        if callback_id == callbacks.after_serverpy_init_app:
            if hasattr(self, "callback_after_serverpy_init_app"):
                self.callback_after_serverpy_init_app(kwargv["scheduler"])

    @classmethod
    def set_current_user_service_data(cls, service_data: dict):
        ServiceEncryptedStorageManager.get_instance().set_current_user_service_data(
            service_id=cls.id, service_data=service_data
        )

    @classmethod
    def update_current_user_service_data(cls, service_data: dict):
        ServiceEncryptedStorageManager.get_instance().update_current_user_service_data(
            service_id=cls.id, service_data=service_data
        )

    @classmethod
    def get_current_user_service_data(cls) -> dict:
        return (
            ServiceEncryptedStorageManager.get_instance().get_current_user_service_data(
                service_id=cls.id
            )
        )

    @classmethod
    def get_blueprint_name(cls):
        return f"{cls.id}_endpoint"

    @classmethod
    def default_address_label(cls):
        # Have to str() it; can't pass a LazyString to json serializer
        return str(_("Reserved for {}").format(cls.name))

    @classmethod
    def reserve_address(cls, wallet: Wallet, address: str, label: str = None):
        # Mark an Address in a persistent way as being reserved by a Service
        if not label:
            label = cls.default_address_label()
        wallet.associate_address_with_service(
            address=address, service_id=cls.id, label=label
        )

    @classmethod
    def reserve_addresses(
        cls,
        wallet: Wallet,
        label: str = None,
        num_addresses: int = 10,
        annotations: dict = None,
    ) -> List[str]:
        """
        Reserve n unused addresses but leave a gap between each one so that this reserved
        range never causes an address gap in the wallet (e.g. if you reserve ten in a
        row it's possible that some wallet software will miss a new tx on the 11th
        address).

        If `label` is not provided, we use Service.default_address_label().

        Optional `annotations` data can be attached to each Address being reserved.
        """
        # Track Service-related addresses in ServiceAnnotationsStorage
        if annotations:
            annotations_storage = ServiceAnnotationsStorage(
                service_id=cls.id, wallet=wallet
            )

        # Start with the addresses that are already reserved but still unused
        addresses: List[str] = wallet.get_associated_addresses(
            service_id=cls.id, unused_only=True
        )
        logger.debug(f"Already have {len(addresses)} addresses reserved for {cls.id}")

        if len(addresses) < num_addresses:
            if addresses:
                # Continuing reserving from where we left off
                index = addresses[-1].index + 2

                # Final `addresses` list has to be just addr strs
                addresses = [addr_obj.address for addr_obj in addresses]
            else:
                index = wallet.address_index + 1

            while len(addresses) < num_addresses:
                address = wallet.get_address(index)
                addr_obj = wallet.get_address_obj(address)

                index += 2

                if addr_obj.used or addr_obj.is_reserved:
                    continue

                # Mark an Address in a persistent way as being reserved by a Service
                cls.reserve_address(wallet=wallet, address=address)
                logger.debug(f"Reserved {address} for {cls.id}")

                addresses.append(address)

                if annotations:
                    annotations_storage.set_addr_annotations(
                        addr=address, annotations=annotations, autosave=False
                    )
        if annotations:
            annotations_storage.save()

        return addresses

    @classmethod
    def unreserve_addresses(cls, wallet: Wallet):
        """
        Clear out Services-related data from any unused Addresses, but leave already-used
        Addresses as-is.
        """
        annotations_storage = ServiceAnnotationsStorage(
            service_id=cls.id, wallet=wallet
        )
        addrs = wallet.get_associated_addresses(service_id=cls.id, unused_only=True)
        for addr_obj in addrs:
            wallet.deassociate_address(addr_obj["address"])
            annotations_storage.remove_addr_annotations(
                addr_obj.address, autosave=False
            )
        annotations_storage.save()

    # def is_active(self):
    #     return self.active

    # def set_active(self, value):
    #     self.active = value

    @classmethod
    def update(self):
        """
        Called by backend periodic process to keep Service in sync with any remote
        data (e.g. fetching the latest data from an external API).
        """
        logger.info(f"update() not implemented / not necessary for Service {self.id}")

    """ ***********************************************************************
                                    Update hooks
    *********************************************************************** """

    @classmethod
    def on_user_login(cls):
        pass

    @classmethod
    def inject_in_basejinja_head(cls):
        """overwrite this method to inject a snippet of code in specter's base.jinja
        the snippet will be placed at the end of the head-section
        a typical implementation would be something like:
        return render_template("myext/inject_in_basejinja_head.jinja")
        """
        pass

    @classmethod
    def inject_in_basejinja_body_top(cls):
        """overwrite this method to inject a snippet of code in specter's base.jinja
        the snippet will be placed at the top of the body-section
        a typical implementation would be something like:
        return render_template("myext/inject_in_basejinja_body_top.jinja")
        """
        pass

    @classmethod
    def inject_in_basejinja_body_bottom(cls):
        """overwrite this method to inject a snippet of code in specter's base.jinja
        the snippet will be placed at the top of the body-section
        a typical implementation would be something like:
        return render_template("myext/inject_in_basejinja_body_bottom.jinja")
        """
        pass



import os
import logging
import secrets
import shutil

from ..rpc import get_default_datadir, RPC_PORTS
from ..specter_error import SpecterError
from ..persistence import write_node, delete_file
from ..helpers import alias, load_jsons
from ..node import Node
from ..internal_node import InternalNode
from ..util.bitcoind_setup_tasks import setup_bitcoind_thread

logger = logging.getLogger(__name__)


class NodeManager:
    # chain is required to manage wallets when bitcoind is not running
    DEFAULT_ALIAS = "default"

    def __init__(
        self,
        proxy_url="socks5h://localhost:9050",
        only_tor=False,
        active_node="default",
        bitcoind_path="",
        internal_bitcoind_version="",
        data_folder="",
    ):
        self.data_folder = data_folder
        self._active_node = active_node
        self.proxy_url = proxy_url
        self.only_tor = only_tor
        self.bitcoind_path = bitcoind_path
        self.internal_bitcoind_version = internal_bitcoind_version
        self.update(data_folder)
        internal_nodes = [
            node for node in self.nodes.values() if not node.external_node
        ]
        for node in internal_nodes:
            node.start()

    def update(self, data_folder=None):
        if data_folder is not None:
            self.data_folder = data_folder
            if data_folder.startswith("~"):
                data_folder = os.path.expanduser(data_folder)
            # creating folders if they don't exist
            if not os.path.isdir(data_folder):
                os.mkdir(data_folder)
        nodes = {}
        nodes_files = load_jsons(self.data_folder, key="name")
        for node_alias in nodes_files:
            fullpath = os.path.join(self.data_folder, "%s.json" % node_alias)
            node_class = (
                Node if nodes_files[node_alias]["external_node"] else InternalNode
            )
            nodes[nodes_files[node_alias]["name"]] = node_class.from_json(
                nodes_files[node_alias],
                self,
                default_alias=node_alias,
                default_fullpath=fullpath,
            )
        if not nodes:
            if os.environ.get("ELM_RPC_USER"):
                self.add_node(
                    node_type="ELM",
                    name="Blockstream Liquid",
                    autodetect=True,
                    datadir=get_default_datadir(node_type="ELM"),
                    user="",
                    password="",
                    port=7041,
                    host="localhost",
                    protocol="http",
                    external_node=True,
                    default_alias=self.DEFAULT_ALIAS,
                )
            self.add_node(
                node_type="BTC",
                name="Bitcoin Core",
                autodetect=True,
                datadir=get_default_datadir(),
                user="",
                password="",
                port=8332,
                host="localhost",
                protocol="http",
                external_node=True,
                default_alias=self.DEFAULT_ALIAS,
            )
        else:
            self.nodes = nodes

    @property
    def active_node(self):
        return self.get_by_alias(self._active_node)

    @property
    def nodes_names(self):
        return sorted(self.nodes.keys())

    def switch_node(self, node_alias):
        # this will throw an error if the node doesn't exist
        self._active_node = self.get_by_alias(node_alias).alias

    def default_node(self):
        return self.get_by_alias(self.DEFAULT_ALIAS)

    def get_by_alias(self, alias):
        for node_name in self.nodes:
            if self.nodes[node_name] and self.nodes[node_name].alias == alias:
                return self.nodes[node_name]
        raise SpecterError("Node %s does not exist!" % alias)

    def update_bitcoind_version(self, specter, version):
        stopped_nodes = []
        for node in (node for node in self.nodes.values() if not node.external_node):
            if node.is_bitcoind_running():
                node.stop()
                stopped_nodes.append(node.alias)
        shutil.rmtree(
            os.path.join(specter.data_folder, "bitcoin-binaries"),
            ignore_errors=True,
        )
        setup_bitcoind_thread(specter, version)
        for node in (node for node in self.nodes.values() if not node.external_node):
            node.version = version
            logger.info(f"persisting {node} in update_bitcoind_version")
            write_node(node, node.fullpath)
        for node_alias in stopped_nodes:
            self.get_by_alias(node_alias).start(timeout=60)

    def add_node(
        self,
        node_type,
        name,
        autodetect,
        datadir,
        user,
        password,
        port,
        host,
        protocol,
        external_node,
        default_alias=None,
    ):
        """Adding a node. Params:
        :param node_type: only valid for autodetect. Either BTC or ELM

        """
        if not default_alias:
            node_alias = alias(name)
        else:
            node_alias = default_alias
        fullpath = os.path.join(self.data_folder, "%s.json" % node_alias)
        i = 2
        while os.path.isfile(fullpath):
            node_alias = alias("%s %d" % (name, i))
            fullpath = os.path.join(self.data_folder, "%s.json" % node_alias)
            i += 1

        node = Node(
            name,
            node_alias,
            autodetect,
            datadir,
            user,
            password,
            port,
            host,
            protocol,
            external_node,
            fullpath,
            node_type,
            self,
        )
        logger.info(f"persisting {node} in add_node")
        write_node(node, fullpath)
        self.update()  # reload files
        logger.info("Added new node {}".format(node.alias))
        return node

    def add_internal_node(
        self,
        name,
        network="main",
        port=None,
        default_alias=None,
        datadir=None,
    ):
        if not default_alias:
            node_alias = alias(name)
        else:
            node_alias = default_alias
        fullpath = os.path.join(self.data_folder, "%s.json" % node_alias)
        i = 2
        while os.path.isfile(fullpath):
            node_alias = alias("%s %d" % (name, i))
            fullpath = os.path.join(self.data_folder, "%s.json" % node_alias)
            i += 1
        if not datadir:
            datadir = os.path.join(self.data_folder, f"{node_alias}/.bitcoin-{network}")

        node = InternalNode(
            name,
            node_alias,
            False,
            datadir,
            "bitcoin",
            secrets.token_urlsafe(16),
            port if port else (RPC_PORTS[network] if network in RPC_PORTS else 8332),
            "localhost",
            "http",
            fullpath,
            self,
            self.bitcoind_path,
            network,
            self.internal_bitcoind_version,
        )
        logger.info(f"persisting {node} in add_internal_node")
        write_node(node, fullpath)
        self.update()  # reload files
        logger.info("Added new internal node {}".format(node.alias))
        return node

    def delete_node(self, node, specter):
        logger.info("Deleting {}".format(node.alias))
        # Delete files
        delete_file(node.fullpath)
        delete_file(node.fullpath + ".bkp")
        if self._active_node == node.alias:
            specter.update_active_node(next(iter(self.nodes.values())).alias)
        del self.nodes[node.name]
        self.update()
        logger.info("Node {} was deleted successfully".format(node.alias))



import logging
import re
from datetime import datetime
import json
from flask_babel.speaklater import LazyString
from typing import Union

logger = logging.getLogger(__name__)


def str2bool(my_str):
    """returns a reasonable boolean from a string so that "False" will result in False"""
    if my_str is None:
        return False
    elif isinstance(my_str, str) and my_str.lower() == "false":
        return False
    elif isinstance(my_str, str) and my_str.lower() == "off":
        return False
    return bool(my_str)


def camelcase2snake_case(name):
    """If you pass DeviceManager it returns device_manager"""
    pattern = re.compile(r"(?<!^)(?=[A-Z])")
    name = pattern.sub("_", name).lower()
    return name


def snake_case2camelcase(word):
    return "".join(x.capitalize() or "_" for x in word.split("_"))


def format_btc_amount_as_sats(
    value: Union[float, str],
    enable_digit_formatting=False,
) -> str:
    s = "{:,.0f}".format(round(float(value) * 1e8))

    # combine the "," with the left number to an array
    array = []
    for letter in s:
        if letter == ",":
            array[-1] += letter
        else:
            array.append(letter)

    if enable_digit_formatting:
        if len(array) >= 4:
            left_index = -6 if len(array) >= 6 else -len(array)
            array[
                left_index
            ] = f'<span class="thousand-digits-in-sats-amount">{array[left_index]}'
            array[-4] = f"{array[-4]}</span>"

        left_index = -3 if len(array) >= 3 else -len(array)
        array[
            left_index
        ] = f'<span class="last-digits-in-sats-amount">{array[left_index]}'
        array[-1] = f"{array[-1]}</span>"

    return "".join(array)


def format_btc_amount(
    value: Union[float, str],
    maximum_digits_to_strip=7,
    minimum_digits_to_strip=6,
    enable_digit_formatting=True,
) -> str:
    """
    Formats the btc amount such that it can be right aligned such
    that the decimal separator will be always at the same x position.

    Stripping trailing 0's is done via just making the 0's transparent.

    Args:
        value (Union[float, str]): Will convert string to float.
            The float is expected to be in the unit (L)BTC with 8 relevant digits
        maximum_digits_to_strip (int, optional): No more than maximum_digits_to_strip
            trailing 0's will be stripped. Defaults to 7.
        minimum_digits_to_strip (int, optional): Only strip any trailing 0's if
            there are at least minimum_digits_to_strip. Defaults to 6.
        enable_digit_formatting (bool, optional): Will group the Satoshis into blocks of 3,
            e.g. 0.03 123 456, and color the blocks. Defaults to True.

    Returns:
        str: The formatted btc amount as html code.
    """
    value = round(float(value), 8)
    formatted_amount = "{:,.8f}".format(value)

    count_digits_that_can_be_stripped = 0
    for i in reversed(range(len(formatted_amount))):
        if formatted_amount[i] == "0":
            count_digits_that_can_be_stripped += 1
            continue
        break

    array = list(formatted_amount)
    if count_digits_that_can_be_stripped >= minimum_digits_to_strip:
        # loop through the float number, e.g. 0.03 000 000, from the right and replace 0's or the '.' until you hit anything != 0
        for i in reversed(range(len(array))):
            if array[i] == "0" and len(array) - i <= maximum_digits_to_strip:
                array[
                    i
                ] = f'<span class="unselectable transparent-text">{array[i]}</span>'
                # since this digit == 0, then continue the loop and check the next digit
                continue
            # the following if branch is only relevant if last_digits_to_strip == 8, i.e. all digits can be stripped
            elif formatted_amount[i] == ".":
                array[
                    i
                ] = f'<span class="unselectable transparent-text">{array[i]}</span>'
                # since this character == '.', then the loop must be broken now
            # always break the loop. Only the digit == 0 can prevent this break
            break

    if enable_digit_formatting:
        array[-6] = f'<span class="thousand-digits-in-btc-amount">{array[-6]}'
        array[-4] = f"{array[-4]}</span>"
        array[-3] = f'<span class="last-digits-in-btc-amount">{array[-3]}'
        array[-1] = f"{array[-1]}</span>"

    return "".join(array)


def robust_json_dumps(obj):
    def default(o):
        if isinstance(o, datetime):
            return o.timestamp()
        if isinstance(o, set):
            return list(o)
        if isinstance(o, LazyString):
            return str(o)
        logger.warning(
            f"robust_json_dumps could not convert {o} of type {type(o)}.  Converting to string instead."
        )
        return str(o)

    return json.dumps(obj, default=default)



from ..services.callbacks import flask_before_request
import random, traceback
from time import time
from flask_wtf.csrf import CSRFError
from werkzeug.exceptions import MethodNotAllowed, NotFound
from flask import render_template, request, redirect, url_for, flash, g
from flask_babel import lazy_gettext as _
from ..specter_error import SpecterError, ExtProcTimeoutException
from pathlib import Path

env_path = Path(".") / ".flaskenv"
from dotenv import load_dotenv

load_dotenv(env_path)

from flask import current_app as app
from .filters import filters_bp

app.register_blueprint(filters_bp)

# Setup specter endpoints
from .welcome import welcome_endpoint
from .auth import auth_endpoint
from .devices import devices_endpoint
from .nodes import nodes_endpoint
from .price import price_endpoint
from .settings import settings_endpoint
from .setup import setup_endpoint
from .wallets import wallets_endpoint
from .wallets_api import wallets_endpoint_api
from ..rpc import RpcError

# Services live in their own separate path
from cryptoadvance.specter.services.controller import services_endpoint

spc_prefix = app.config["SPECTER_URL_PREFIX"]
app.register_blueprint(welcome_endpoint, url_prefix=f"{spc_prefix}/welcome")
app.register_blueprint(auth_endpoint, url_prefix=f"{spc_prefix}/auth")
app.register_blueprint(devices_endpoint, url_prefix=f"{spc_prefix}/devices")
app.register_blueprint(nodes_endpoint, url_prefix=f"{spc_prefix}/nodes")
app.register_blueprint(price_endpoint, url_prefix=f"{spc_prefix}/price")
app.register_blueprint(services_endpoint, url_prefix=f"{spc_prefix}/services")
app.register_blueprint(settings_endpoint, url_prefix=f"{spc_prefix}/settings")
app.register_blueprint(setup_endpoint, url_prefix=f"{spc_prefix}/setup")
app.register_blueprint(wallets_endpoint, url_prefix=f"{spc_prefix}/wallets")
app.register_blueprint(wallets_endpoint_api, url_prefix=f"{spc_prefix}/wallets")

rand = random.randint(0, 1e32)  # to force style refresh

########## exception handlers ##############
@app.errorhandler(RpcError)
def server_rpc_error(rpce):
    """Specific SpecterErrors get passed on to the User as flash"""
    if rpce.error_code == -18:  # RPC_WALLET_NOT_FOUND
        flash(
            _("Wallet not found. Specter reloaded all wallets, please try again."),
            "error",
        )
    else:
        flash(_("Bitcoin Core RpcError: {}").format(str(rpce)), "error")
    try:
        app.specter.wallet_manager.update()
    except SpecterError as se:
        flash(str(se), "error")
    return redirect(url_for("welcome_endpoint.about"))


@app.errorhandler(SpecterError)
def server_specter_error(se):
    """Specific SpecterErrors get passed on to the User as flash"""
    flash(str(se), "error")
    try:
        app.specter.wallet_manager.update()
    except SpecterError as se:
        flash(str(se), "error")
    if request.method == "POST":
        return redirect(request.url)
    # potentially avoiding http loops. Might be improvable but how?
    else:
        return redirect(url_for("welcome_endpoint.about"))


@app.errorhandler(NotFound)
def server_notFound_error(e):
    """Unspecific Exceptions get a 404 Error-Page"""
    # if rpc is not available
    app.logger.error("Could not find Resource (404): %s" % request.url)
    return render_template("500.jinja", error=e), 404


@app.errorhandler(Exception)
def server_error(e):
    """Unspecific Exceptions get a 500 Error-Page"""
    # if rpc is not available
    if app.specter.rpc is None or not app.specter.rpc.test_connection():
        # make sure specter knows that rpc is not there
        app.specter.check()
    app.logger.error("Uncaught exception: %s" % e)
    trace = traceback.format_exc()
    app.logger.error(trace)
    return render_template("500.jinja", error=e, traceback=trace), 500


@app.errorhandler(ExtProcTimeoutException)
def server_error_timeout(e):
    """Unspecific Exceptions get a 500 Error-Page"""
    # if rpc is not available
    if app.specter.rpc is None or not app.specter.rpc.test_connection():
        # make sure specter knows that rpc is not there
        app.specter.check()
    app.logger.error("ExternalProcessTimeoutException: %s" % e)
    flash(
        _(
            "Bitcoin Core is not coming up in time. Maybe it's just slow but please check the logs below"
        ),
        "warn",
    )
    return redirect(
        url_for(
            "node_settings.bitcoin_core_internal_logs",
            node_alias=app.specter.node.alias,
        )
    )


@app.errorhandler(CSRFError)
def server_error_csrf(e):
    """CSRF token missing. Most likely session expired.
    If persisting after refresh this could mean the front-end
    is not sending the CSRF token properly in some form"""
    app.logger.error("CSRF Exception: %s" % e)
    trace = traceback.format_exc()
    app.logger.error(trace)
    flash(_("Session expired. Please refresh and try again."), "error")
    return redirect(request.url)


@app.errorhandler(MethodNotAllowed)
def server_error_405(e):
    """405 method not allowed. Token might have expired."""
    app.logger.error("405 MethodNotAllowed Exception: %s" % e)
    trace = traceback.format_exc()
    app.logger.error(trace)
    flash(_("Session expired. Please refresh and try again."), "error")
    return redirect(request.url)


########## on every request ###############
@app.before_request
def selfcheck():
    """check status before every request"""
    if app.specter.rpc is not None:
        type(app.specter.rpc).counter = 0
        if not app.specter.chain:
            app.specter.check()
    if app.config.get("LOGIN_DISABLED"):
        app.login("admin")


@app.before_request
def slow_request_detection_start():
    g.start = time()


@app.before_request
def execute_service_manager_hook():
    """inform extensions about the request"""
    app.specter.service_manager.execute_ext_callbacks(flask_before_request, request)


@app.after_request
def slow_request_detection_stop(response):
    try:
        diff = time() - g.start
    except Exception as e:
        app.logger.error(e)
        return response
    if (
        (response.response)
        and (200 <= response.status_code < 300)
        and (response.content_type.startswith("text/html"))
    ):
        threshold = app.config["REQUEST_TIME_WARNING_THRESHOLD"]
        if diff > threshold:
            flash(
                _(
                    "The request before this one took {} seconds which is longer than the threshold ({}). Checkout the perfomance-improvement-hints in the documentation".format(
                        int(diff), threshold
                    )
                ),
                "warning",
            )
    return response


########## template injections #############
@app.context_processor
def inject_common_stuff():
    """Can be used in all jinja2 templates"""
    return dict(
        debug=app.config["DEBUG"],
        specter_url_prefix=app.config["APP_URL_PREFIX"]
        + app.config["SPECTER_URL_PREFIX"],
        ext_url_prefix=app.config["APP_URL_PREFIX"] + app.config["EXT_URL_PREFIX"],
    )


################ Specter global routes ####################
@app.route("/")
def index():
    if app.config["SPECTER_URL_PREFIX"] == "":
        return redirect(url_for("welcome_endpoint.index"))
    else:
        """This is the root-entry URL which redirects to ROOT_URL_REDIRECT"""
        return redirect(app.config["ROOT_URL_REDIRECT"])


if app.config["SPECTER_URL_PREFIX"] != "":
    # Not necessary if the prefix has been removed
    @app.route(f"{app.config['SPECTER_URL_PREFIX']}/")
    def index_prefix():
        return redirect(url_for("welcome_endpoint.index"))


@app.route("/healthz/liveness")
def liveness():
    return {"message": "i am alive"}


@app.route("/healthz/readyness")
def readyness():
    try:
        # Probably improvable:
        app.specter.check()
    except Exception as e:
        return {"message": "i am not ready"}, 500
    return {"message": "i am ready"}



import logging
import multiprocessing
import os
import signal
import sys
import time
from os import path, name
from socket import gethostname
from urllib.parse import urlparse

import click

try:
    from cryptoadvance.specter.gunicorn import SpecterGunicornApp
except ModuleNotFoundError as e:
    # gunicorn is not supported on windows
    if os.name == "nt":
        pass
    else:
        raise e
from cryptoadvance.specter.server import create_and_init, create_app, init_app
from OpenSSL import SSL, crypto
from stem.control import Controller

from ..server import create_app, init_app
from ..specter_error import SpecterError
from ..util.tor import start_hidden_service, stop_hidden_services

logger = logging.getLogger(__name__)


@click.group()
def cli():
    pass


@cli.command()
def gunicorn():
    # fmt: off
    """uses gunicorn instead of the build in development-server.
    
    This works mostly like in https://docs.gunicorn.org/en/latest/run.html just that you use
    this command rather than gunicorn. As recommended there, everything Application
    specific needs to be configured via envirnoment Vars. See cryptoadvance.specter.config for
    how to configure most specter-sepcific things.

    Other than running gunicorn's executable directly, you can't specify command-line
    paramaters. Here are some configurations you can do:

    \b
    GUNICORN_CMD_ARGS="--bind=127.0.0.1 --workers=3" \ 
      python3 -m cryptoadvance.specter gunicorn
    
    You can use a config file named gunicorn.conf.py in the same directory:

    \b
    workers=10

    Commandline paramater trums env-var params.
    You can specify hook functions in the gunicorn.conf.py, see this for a list of them:
    https://docs.gunicorn.org/en/latest/settings.html#server-hooks

    e.g.:

    \b
    def on_starting(server):
        print("Called just before the master process is initialized.")
    
    More information directly in the gunicorn-Documentation

    """
    if os.name == "nt":
        print("Sorry, gunicorn is not available on windows")
        exit(1)
    specter_gunicorn = SpecterGunicornApp(config=None)
    specter_gunicorn.run()



""" Here we have some constants getting an id for extension-points/callbacks. As camelcase
    is used, we don't use CAPITAL letters to not loose the meaning of the camelcase.
    
    These constants are expected as parameter to the ServiceManager.callback function
    and it'll throw an exception if the constant does not exist.

    There are some weak naming conventions:
    1. after/before
    2. class or file
    3. method or function
    
"""

"""
    I don't know why we have this one. Doesn't seem to be used anywhere.    
"""
afterServiceManagerInit = "afterServiceManagerInit"

"""
    This one is called, after the init_app method has finished. The "run" method has not 
    been executed yet and so urls can't be called yet.
    So this is the best place for almost all extensions to do their initializing work.
"""
after_serverpy_init_app = "after_serverpy_init_app"

"""
    This one is called, whenever a file is persisted. To call external scripts in another
    process, you better use the SPECTER_PERSISTENCE_CALLBACK Env Var or it's asynchronous cousin
    SPECTER_PERSISTENCE_CALLBACK_ASYNC.
"""
specter_persistence_callback = "specter_persistence_callback"

""" 
    Will get called before every request via the Flask's @app.before_request
"""
flask_before_request = "flask_before_request"



import logging
import os
import shutil
import signal
import sys
import time
from datetime import datetime
from os import path
from os.path import exists, getmtime, join
from pathlib import Path
from socket import gethostname
from sre_constants import BRANCH
from urllib.parse import urlparse

import click
import requests
from jinja2 import BaseLoader, Environment, FileSystemLoader, TemplateNotFound

from ..server import create_app, init_app
from ..specter_error import SpecterError
from ..util.common import camelcase2snake_case, snake_case2camelcase
from ..util.reflection_fs import search_dirs_in_path
from ..util.shell import run_shell
from ..util.tor import start_hidden_service, stop_hidden_services

ext_mark = "specterext"
logger = logging.getLogger(__name__)


class ExtGen:
    def __init__(
        self,
        base_path,
        ext_org,
        ext_id,
        isolated_client,
        author,
        email,
        dry_run=False,
        branch="master",
        tmpl_fs_source=None,
    ):
        self.base_path = base_path
        self.org = ext_org
        self.id = ext_id
        self.isolated_client = isolated_client
        self.author = author
        self.author_email = email
        self.version = "1.8.1"  # relevant if tmpl-sources specify a dependency (requirements.txt) #ToDo improve
        self.branch = branch
        self.tmpl_fs_source = tmpl_fs_source

        self.dry_run = dry_run
        self.create_envs()

    def create_envs(self):
        if self.tmpl_fs_source == None:
            loader = GithubUrlLoader(branch=self.branch)
        else:
            loader = FileSystemLoader(self.tmpl_fs_source)
        self.jinja_env = Environment(
            loader=loader,
            trim_blocks=True,
            block_start_string="<<",
            block_end_string=">>",
            variable_start_string="<=",
            variable_end_string="=>",
            comment_start_string="<#",
            comment_end_string="#>",
        )
        self.jinja_env.filters["camelcase"] = snake_case2camelcase
        self.env = Environment(loader=loader, trim_blocks=True)
        self.env.filters["camelcase"] = snake_case2camelcase
        self.sd_env = Environment(
            loader=GithubUrlLoader(
                base_url="https://raw.githubusercontent.com/cryptoadvance/specter-desktop/",
                branch=self.branch,
            )
        )

    def env_for_template(self, template):
        """chooses the right env for the template"""
        if Path(template).name in ["conftest.py", "fix_ghost_machine.py"]:
            return self.sd_env
        if Path(template).suffix.endswith("jinja"):
            return self.jinja_env
        return self.env

    def generate(self):
        self.generate_basics()
        self.generate_preppub()

    def generate_basics(self):
        self.render("requirements.txt", version=self.version)
        self.render(".gitignore")
        package_path = f"src/dummyorg/specterext/dummy"
        self.render(f"{package_path}/service.py")
        self.render(f"{package_path}/controller.py")
        self.render(f"{package_path}/config.py")
        self.render(f"{package_path}/app_config.py")
        self.render(f"{package_path}/__init__.py")
        self.render(f"{package_path}/__main__.py")
        self.render(f"{package_path}/templates/dummy/index.jinja")
        if not self.isolated_client:
            self.render(f"{package_path}/static/dummy/css/styles.css")
            self.create_binary_file(f"{package_path}/static/dummy/img/ghost.png")
            self.create_binary_file(f"{package_path}/static/dummy/img/logo.jpeg")
            self.render(f"{package_path}/templates/dummy/base.jinja")
            self.render(f"{package_path}/templates/dummy/transactions.jinja")
            self.render(f"{package_path}/templates/dummy/settings.jinja")
            self.render(f"{package_path}/templates/dummy/components/dummy_menu.jinja")
            self.render(f"{package_path}/templates/dummy/components/dummy_tab.jinja")

        self.render(f"pytest.ini", env=self.sd_env)
        self.render(f"tests/conftest.py", env=self.sd_env)
        self.render(f"tests/fix_ghost_machine.py", env=self.sd_env)
        self.render(f"tests/fix_devices_and_wallets.py", env=self.sd_env)
        self.render(f"tests/fix_testnet.py", env=self.sd_env)
        self.render(f"tests/fix_keys_and_seeds.py", env=self.sd_env)

    def create_binary_file(self, sourcepath):
        """textfiles can all be rendered. Binaries must be wgettet or copied"""
        targetpath = Path(
            sourcepath.replace("dummyorg", self.org).replace("dummy", self.id)
        )
        if self.dry_run:
            print(f"-------------------------------------")
            print(" Creation of targetpath skipped because dryrun")
            print(f"-------------------------------------")
            return
        if targetpath.is_file():
            return
        targetpath.parents[0].mkdir(parents=True, exist_ok=True)
        if self.tmpl_fs_source != None:
            sourcepath = Path(self.tmpl_fs_source, sourcepath)
            shutil.copy(sourcepath, targetpath)
            print(f"    --> Created {targetpath} (copied)")
        else:
            r = requests.get(self.env.loader.url_for_template(sourcepath))
            open(targetpath, "wb").write(r.content)
            print(f"    --> Created {targetpath} (via Github)")

    def generate_preppub(self):
        self.render("pyproject.toml", version=self.version)
        self.render("setup.py")
        # Author and Email
        if not self.author:
            result = run_shell(["git", "config", "--get", "user.name"])
            if result["code"] == 0:
                author = result["out"].decode("ascii").strip()
            else:
                author = click.prompt("Please type in your Name: ", type=str)
        if not self.author_email:
            result = run_shell(["git", "config", "--get", "user.email"])
            if result["code"] == 0:
                email = result["out"].decode("ascii").strip()
            else:
                email = click.prompt("Please type in your E-Mail: ", type=str)
        self.render("setup.cfg")
        self.render("MANIFEST.in")

    def render(self, template, env=None, **kargv):
        # The template path is the same that we want to store on disk
        file_name = Path(
            template.replace("dummyorg", self.org).replace("dummy", self.id)
        )
        if env == None:
            env = self.env_for_template(template)
        template = env.get_template(template)

        if not self.dry_run:
            file_name.parents[0].mkdir(parents=True, exist_ok=True)
        rendered_text = template.render(ext=self, **kargv)
        if self.dry_run:
            print()
            print(
                f"File: {file_name} ({env.block_start_string} {env.block_end_string})"
            )
            # print(f"Url: ")
            print("-------------------------------------------------")
            print(rendered_text)
            print("-------------------------------------------------")
            print()
        else:
            fq_fname = Path(self.base_path, file_name)
            if not fq_fname.is_file():
                with open(fq_fname, "w") as file:
                    file.write(rendered_text)
                    print(f"    --> Created {fq_fname}")


class GithubUrlLoader(BaseLoader):
    """A Jinja2 TemplateLoader which is loading templates directly from github.
    If you don't specify it, it'll be https://github.com/cryptoadvance/specterext-dummy
    """

    dummy_base_url = "https://raw.githubusercontent.com/cryptoadvance/specterext-dummy"

    def __init__(self, base_url=None, branch=None):
        if base_url == None:
            base_url = self.dummy_base_url
        if branch == None:
            branch = "master"
        self.base_url = base_url + "/" + branch

    def url_for_template(self, template):
        if template[0] == "/":
            template = template[1:]
        url = self.base_url + "/" + template
        return url

    def get_source(self, environment, template):
        url = self.url_for_template(template)
        r = requests.get(url)
        if r.status_code != 200:
            raise TemplateNotFound(f"{url} results in {r.status_code}")
        source = r.text
        return source, url, None



import datetime
import json
import logging
import pytz

from flask import current_app as app, request
from flask_babel import lazy_gettext as _
from typing import List
from cryptoadvance.specter.services.swan.client import SwanClient
from cryptoadvance.specter.specter_error import SpecterError

from cryptoadvance.specter.user import User

from ..service import Service, devstatus_prod
from cryptoadvance.specter.addresslist import Address
from cryptoadvance.specter.wallet import Wallet
from urllib.parse import urlparse

logger = logging.getLogger(__name__)


class SwanService(Service):
    id = "swan"
    name = "Swan"
    icon = "swan/img/swan_icon.svg"
    logo = "swan/img/swan_logo.svg"
    desc = "Auto-withdraw to your Specter wallet"
    has_blueprint = True
    isolated_client = False
    devstatus = devstatus_prod

    # TODO: As more Services are integrated, we'll want more robust categorization and sorting logic
    sort_priority = 1

    # Service-specific constants
    MIN_PENDING_AUTOWITHDRAWAL_ADDRS = 10

    # ServiceEncryptedStorage field names for Swan
    SPECTER_WALLET_ALIAS = "wallet"
    SWAN_WALLET_ID = "swan_wallet_id"
    ACCESS_TOKEN = "access_token"
    ACCESS_TOKEN_EXPIRES = "expires"
    REFRESH_TOKEN = "refresh_token"
    AUTOWITHDRAWAL_ID = "autowithdrawal_id"
    AUTOWITHDRAWAL_THRESHOLD = "withdrawal_threshold"

    @classmethod
    def client(cls) -> SwanClient:
        if hasattr(cls, "_client"):
            return cls._client
        try:
            cls._client = SwanClient(
                urlparse(request.url).netloc,
                cls.get_current_user_service_data().get(cls.ACCESS_TOKEN),
                cls.get_current_user_service_data().get(cls.ACCESS_TOKEN_EXPIRES),
                cls.get_current_user_service_data().get(cls.REFRESH_TOKEN),
            )
        except Exception as e:
            raise e
        return cls._client

    @classmethod
    def is_access_token_valid(cls):
        service_data = cls.get_current_user_service_data()
        if not service_data or not service_data.get(cls.ACCESS_TOKEN_EXPIRES):
            return False
        return (
            service_data[cls.ACCESS_TOKEN_EXPIRES]
            > datetime.datetime.now(tz=pytz.utc).timestamp()
        )

    @classmethod
    def store_new_api_access_data(cls):
        new_api_data = {
            cls.ACCESS_TOKEN: cls.client().access_token,
            cls.ACCESS_TOKEN_EXPIRES: cls.client().access_token_expires,
        }
        if cls.client().refresh_token:
            new_api_data[cls.REFRESH_TOKEN] = cls.client().refresh_token
        logger.debug(f"Storing: {new_api_data}")
        cls.update_current_user_service_data(new_api_data)

    @classmethod
    def has_refresh_token(cls):
        return cls.REFRESH_TOKEN in cls.get_current_user_service_data()

    @classmethod
    def get_associated_wallet(cls) -> Wallet:
        """Get the Specter `Wallet` that is currently associated with Swan auto-withdrawals"""
        service_data = cls.get_current_user_service_data()
        if not service_data or cls.SPECTER_WALLET_ALIAS not in service_data:
            # Service is not initialized; nothing to do
            return
        try:
            return app.specter.wallet_manager.get_by_alias(
                service_data[cls.SPECTER_WALLET_ALIAS]
            )
        except SpecterError as e:
            logger.debug(e)
            # Referenced an unknown wallet
            # TODO: keep ignoring or remove the unknown wallet from service_data?
            return

    @classmethod
    def set_associated_wallet(cls, wallet: Wallet):
        """Set the Specter `Wallet` that is currently associated with Swan auto-withdrawals"""
        cls.update_current_user_service_data({cls.SPECTER_WALLET_ALIAS: wallet.alias})

    @classmethod
    def reserve_addresses(
        cls, wallet: Wallet, label: str = None, num_addresses: int = 10
    ) -> List[str]:
        """
        * Reserves addresses for Swan auto-withdrawals
        * Sets the associated Specter `Wallet` that will receive auto-withdrawals
        * Removes any existing unused reserved addresses in the previously associated `Wallet`
        * Performs matching cleanup and update on the Swan side

        Overrides base classmethod to add Swan-specific functionality & data management.
        """
        from . import client as swan_client

        # Update Addresses as reserved (aka "associated") with Swan in our Wallet
        addresses = super().reserve_addresses(
            wallet=wallet, label=label, num_addresses=num_addresses
        )

        # Clear out any prior unused reserved addresses if this is a different Wallet
        cur_wallet = cls.get_associated_wallet()
        if cur_wallet and cur_wallet != wallet:
            super().unreserve_addresses(cur_wallet)

        # Store our `Wallet` as the current one for Swan auto-withdrawals
        cls.set_associated_wallet(wallet)

        # Send the new list to Swan (DELETES any unused ones; creates a new SWAN_WALLET_ID if needed)
        swan_wallet_id = cls.client().update_autowithdrawal_addresses(
            cls.get_current_user_service_data().get(cls.SWAN_WALLET_ID),
            specter_wallet_name=wallet.name,
            specter_wallet_alias=wallet.alias,
            addresses=addresses,
        )
        logger.debug(f"Updating the Swan wallet id to {swan_wallet_id}")
        if swan_wallet_id:
            cls.update_current_user_service_data({cls.SWAN_WALLET_ID: swan_wallet_id})

        return addresses

    @classmethod
    def set_autowithdrawal_settings(cls, wallet: Wallet, btc_threshold: str):
        """
        btc_threshold: "0", "0.01", "0.025", or "0.05"

        Performs a lot of maintenance behind the scenes in order to keep Specter's
        internal data in sync (e.g. resetting previously reserved addresses) and the same
        in the api to keep Swan's notion of a wallet and list of addrs in sync.
        """
        from . import client as swan_client

        # Reserve auto-withdrawal addresses for this Wallet; clear out an unused ones in a prior wallet
        cls.reserve_addresses(
            wallet=wallet, num_addresses=cls.MIN_PENDING_AUTOWITHDRAWAL_ADDRS
        )

        swan_wallet_id = cls.get_current_user_service_data().get(cls.SWAN_WALLET_ID)
        # Send the autowithdrawal threshold
        resp = cls.client().set_autowithdrawal(
            swan_wallet_id, btc_threshold=btc_threshold
        )
        autowithdrawal_id = resp["item"]["id"]
        if autowithdrawal_id != cls.get_current_user_service_data().get(
            SwanService.AUTOWITHDRAWAL_ID
        ):
            cls.update_current_user_service_data(
                {
                    SwanService.AUTOWITHDRAWAL_ID: autowithdrawal_id,
                }
            )

        # Store the threshold setting in the User's service data
        cls.update_current_user_service_data(
            {SwanService.AUTOWITHDRAWAL_THRESHOLD: btc_threshold}
        )

    @classmethod
    def sync_swan_data(cls):
        """
        Called when the user completes the OAuth2 link with Swan.

        User could be:
        * A first-time Specter-Swan integration (so no local or Swan API wallet data).
        * Re-linking on a previously linked Specter instance (some/all existing data).
        * Linking a new Specter instance but had previously linked on a different Specter
            instance; the previously linked Specter wallet may or may not be present
            (need to resync data).
        """
        from . import client as swan_client

        def sync_autowithdrawal_settings(service_data):
            """
            Retrieve the autowithdrawal objs from Swan. We only care if we find one
            that matches our SWAN_WALLET_ID and has `isActive: true`.

            Otherwise clear any local autowithdrawal data.
            """
            autowithdrawal_info = cls.client().get_autowithdrawal_info()
            """
                {
                    "entity": "automaticWithdrawal",
                    "list": [
                        {
                            "id": "2026d75b-baf0-45c3-a1c6-9e17d4f7e90f",
                            "minBtcThreshold": "0.01",
                            "isActive": false,
                            "isCanceled": false,
                            "createdAt": "2022-01-07T02:14:56.070Z",
                            "walletId": "e26132e7-2e03-49b7-807b-9067aa3a8507",
                            "walletAddressId": null
                        },
                        ...,
                        { ... }
                    ]
                }
            """
            if autowithdrawal_info and "list" in autowithdrawal_info:
                swan_wallet_id = service_data.get(SwanService.SWAN_WALLET_ID)
                for entry in autowithdrawal_info["list"]:
                    if entry.get("wallet_id") == swan_wallet_id and entry.get(
                        "isActive"
                    ):
                        # Found our swan_wallet_id's active autowithdrawal!
                        logger.debug(
                            "Found swan_wallet_id's active autowithdrawal entry"
                        )
                        SwanService.update_current_user_service_data(
                            {
                                SwanService.AUTOWITHDRAWAL_ID: entry["id"],
                                SwanService.AUTOWITHDRAWAL_THRESHOLD: entry[
                                    "minBtcThreshold"
                                ],
                            }
                        )

                        # Check to make sure we have enough autowithdrawal addresses
                        SwanService.update()

                        return

            # Did not find a matching and/or active autowithdrawal
            logger.debug("No active autowithdrawal with swan_wallet_id found")

            # Clear the autowithdrawal fields from local storage and move on
            service_data.pop(SwanService.AUTOWITHDRAWAL_ID, None)
            service_data.pop(SwanService.AUTOWITHDRAWAL_THRESHOLD, None)
            SwanService.set_current_user_service_data(service_data)
            return

        service_data = cls.get_current_user_service_data()
        if SwanService.SWAN_WALLET_ID in service_data:
            # This user has previously/currently linked to Swan on this instance
            swan_wallet_id = service_data[SwanService.SWAN_WALLET_ID]
            logger.debug(f"swan_wallet_id: {swan_wallet_id}")

            # Confirm that the Swan walletId exists on the Swan side
            details = cls.client().get_wallet_details(swan_wallet_id)
            """
            {
                "entity": "wallet",
                "item": {
                    "id": "*******",
                    "isConfirmed": false,
                    "displayName": "Specter autowithdrawal to SeedSigner demo",
                    "metadata": {
                        "oidc": {
                            "clientId": "specter-dev"
                        },
                        "specter_wallet_alias": "seedsigner_demo"
                    }
                }
            }
            """
            if not details:
                # Specter's SWAN_WALLET_ID is out of sync; doesn't exist on Swan's side.
                # Clear the local SWAN_WALLET_ID and continue below to try to find one.
                logger.debug(f"swan_wallet_id {swan_wallet_id} not found on Swan")
                del service_data[SwanService.SWAN_WALLET_ID]
                cls.set_current_user_service_data(service_data)

            elif (
                "item" in details
                and "metadata" in details["item"]
                and "specter_wallet_alias" in details["item"]["metadata"]
            ):
                wallet_alias = details["item"]["metadata"]["specter_wallet_alias"]
                logger.debug(f"swan_wallet_id exists on Swan side")
                if wallet_alias in [
                    w.alias for w in app.specter.wallet_manager.wallets.values()
                ]:
                    # All is good; we've matched Swan's wallet data with a Specter `Wallet` that we recognize.
                    logger.debug(f"Found wallet_alias {wallet_alias} in Specter")
                    if wallet_alias != service_data.get(
                        SwanService.SPECTER_WALLET_ALIAS
                    ):
                        # Our local `service_data` is out of sync; update with the
                        # current Specter wallet_alias Swan is expecting.
                        logger.debug(
                            f"Updating service_data to use wallet_alias {wallet_alias}"
                        )
                        cls.update_current_user_service_data(
                            {SwanService.SPECTER_WALLET_ALIAS: wallet_alias}
                        )

                    sync_autowithdrawal_settings(service_data)

                    return
                else:
                    # Swan is out of sync with Specter; the SPECTER_WALLET_ALIAS we had
                    # been using doesn't exist on this Specter instance.
                    logger.warn(
                        f"Swan referenced an unknown wallet_alias {wallet_alias}"
                    )
                    if SwanService.SPECTER_WALLET_ALIAS in service_data:
                        # Clear the local reference to that unknown SPECTER_WALLET_ALIAS
                        del service_data[SwanService.SPECTER_WALLET_ALIAS]
                        cls.set_current_user_service_data(service_data)

        # This Specter instance has no idea if there might already be wallet data on the Swan side.
        # Fetch all Swan wallets, if any exist.
        wallet_entries = cls.client().get_wallets().get("list")
        if not wallet_entries:
            # No Swan data at all yet. Nothing to do.
            logger.debug("No wallets on the Swan side yet")
            return

        swan_wallet_id = None
        for wallet_entry in wallet_entries:
            swan_wallet_id = wallet_entry["id"]
            specter_wallet_alias = wallet_entry["metadata"].get("specter_wallet_alias")
            if specter_wallet_alias in [
                w.alias for w in app.specter.wallet_manager.wallets.values()
            ]:
                # All is good; we've matched Swan's wallet data with a Specter `Wallet` that we recognize.
                # Use this Swan walletId going forward.
                cls.update_current_user_service_data(
                    {
                        SwanService.SWAN_WALLET_ID: swan_wallet_id,
                        SwanService.SPECTER_WALLET_ALIAS: specter_wallet_alias,
                    }
                )
                logger.debug(
                    f"Found a Specter wallet that we recognize: {specter_wallet_alias}"
                )

                sync_autowithdrawal_settings(service_data)

                return

        # We didn't find a matching Specter `Wallet`. Clear out any nonsense settings in our local data.
        logger.debug(
            "Did not find any matching Specter wallets in the Swan wallet metadata"
        )
        if SwanService.SPECTER_WALLET_ALIAS in service_data:
            del service_data[SwanService.SPECTER_WALLET_ALIAS]
            SwanService.set_current_user_service_data(service_data)

        # Did we at least get a Swan walletId that we can update later?
        if swan_wallet_id:
            cls.update_current_user_service_data(
                {
                    SwanService.SWAN_WALLET_ID: swan_wallet_id,
                }
            )
            logger.debug(f"Setting swan_wallet_id to {swan_wallet_id}")

    @classmethod
    def remove_swan_integration(cls, user: User):
        # Unreserve unused addresses in all wallets
        for wallet_name, wallet in user.wallet_manager.wallets.items():
            SwanService.unreserve_addresses(wallet=wallet)

        # If an autowithdrawal setup is active, remove pending addrs from Swan
        try:
            service_data = SwanService.get_current_user_service_data()
            if service_data.get(cls.SPECTER_WALLET_ALIAS) and service_data.get(
                cls.SWAN_WALLET_ID
            ):
                # Import here to prevent circular dependency
                from . import client as swan_client

                cls.client().delete_autowithdrawal_addresses(
                    service_data[cls.SWAN_WALLET_ID]
                )
        except Exception as e:
            # Note the exception but proceed with clearing local data
            logger.exception(e)

        # Wipe the on-disk encrypted service data (refresh_token, etc)
        SwanService.set_current_user_service_data({})

        # Remove Swan from User's list of active Services
        user.remove_service(SwanService.id)

    """ ***********************************************************************
                                Update hooks overrides
    *********************************************************************** """

    @classmethod
    def update(cls):
        """
        Periodic or at-login call to check our Swan address status and send more when
        needed.
        * Check for autowithdrawals paid to addrs reserved for Swan.
        * Add more pending autowithdrawal addrs if we're under the threshold.
        """
        # Which Specter `Wallet` has been configured to receive Swan autowithdrawals?
        wallet = cls.get_associated_wallet()
        if not wallet:
            # Swan autowithdrawals to Specter aren't set up yet; nothing to do.
            logger.debug(
                "No associated wallet. Swan autowithdrawals are not set up yet."
            )
            return

        # Scan the Wallet for any new Swan autowithdrawals
        reserved_addresses: List[Address] = wallet.get_associated_addresses(
            service_id=cls.id, unused_only=False
        )
        for addr_obj in reserved_addresses:
            if addr_obj.used and addr_obj.label == cls.default_address_label():
                # This addr has received an autowithdrawal since we last checked
                logger.debug(
                    f"Updating address label for {json.dumps(addr_obj, indent=4)}"
                )
                wallet.setlabel(addr_obj.address, str(_("Swan autowithdrawal")))

        num_pending_autowithdrawal_addrs = len(
            [addr_obj for addr_obj in reserved_addresses if not addr_obj["used"]]
        )
        if num_pending_autowithdrawal_addrs < cls.MIN_PENDING_AUTOWITHDRAWAL_ADDRS:
            logger.debug("Need to send more addrs to Swan")
            cls.reserve_addresses(
                wallet=wallet, num_addresses=cls.MIN_PENDING_AUTOWITHDRAWAL_ADDRS
            )

    @classmethod
    def on_user_login(cls):
        cls.update()



import logging
import os
import psutil
from pathlib import Path


from .helpers import is_testnet
from .specter_error import SpecterError, ExtProcTimeoutException
from .rpc import (
    get_default_datadir,
)
from .process_controller.bitcoind_controller import BitcoindPlainController
from .persistence import write_node
from .node import Node

logger = logging.getLogger(__name__)


class InternalNode(Node):
    """A Node but other than Node, this one is managed by Specter.
    So it has start and stop methods and one called is_bitcoind_running
    """

    # Possible Stati
    BROKEN = "Broken"
    DOWN = "Down"
    RUNNING = "Running"

    def __init__(
        self,
        name,
        alias,
        autodetect,
        datadir,
        user,
        password,
        port,
        host,
        protocol,
        fullpath,
        manager,
        bitcoind_path,
        bitcoind_network,
        version,
    ):
        super().__init__(
            name,
            alias,
            autodetect,
            datadir,
            user,
            password,
            port,
            host,
            protocol,
            False,
            fullpath,
            "BTC",
            manager,
        )

        self.bitcoind_path = bitcoind_path
        self.bitcoind_network = bitcoind_network
        self._bitcoind = None
        self.version = version
        self._status = self.DOWN
        if self.bitcoind_network != "main":
            if self.bitcoind_network == "testnet" and not self.datadir.endswith(
                "testnet3"
            ):
                self.datadir = os.path.join(self.datadir, "testnet3")
            elif self.bitcoind_network == "regtest" and not self.datadir.endswith(
                "regtest"
            ):
                self.datadir = os.path.join(self.datadir, "regtest")
            elif self.bitcoind_network == "signet" and not self.datadir.endswith(
                "signet"
            ):
                self.datadir = os.path.join(self.datadir, "signet")
            logger.info(f"persisting {self} in __init__")
            write_node(self, self.fullpath)
        Path(self.datadir).mkdir(parents=True, exist_ok=True)

    @classmethod
    def from_json(cls, node_dict, manager, default_alias="", default_fullpath=""):
        name = node_dict.get("name", "")
        alias = node_dict.get("alias", default_alias)
        autodetect = node_dict.get("autodetect", True)
        datadir = node_dict.get("datadir", get_default_datadir())
        user = node_dict.get("user", "")
        password = node_dict.get("password", "")
        port = node_dict.get("port", None)
        host = node_dict.get("host", "localhost")
        protocol = node_dict.get("protocol", "http")
        external_node = node_dict.get("external_node", True)
        fullpath = node_dict.get("fullpath", default_fullpath)
        bitcoind_path = node_dict.get("bitcoind_path", "")
        bitcoind_network = node_dict.get("bitcoind_network", "main")
        version = node_dict.get("version", "")

        return cls(
            name,
            alias,
            autodetect,
            datadir,
            user,
            password,
            port,
            host,
            protocol,
            fullpath,
            manager,
            bitcoind_path,
            bitcoind_network,
            version,
        )

    @property
    def json(self):
        node_json = super().json
        node_json["bitcoind_path"] = self.bitcoind_path
        node_json["bitcoind_network"] = self.bitcoind_network
        node_json["version"] = self.version
        return node_json

    def start(self, timeout=15):
        """Failsafe way to start an internal node."""
        potential_bitcoind_process = BitcoindProcess.by_password(self.password)
        if potential_bitcoind_process:
            logger.info(
                f"Skipping start of internal node, existing one found {potential_bitcoind_process}"
            )
            self.bitcoind.attach_to_proc_id(potential_bitcoind_process)
            self._status = self.RUNNING
            return self.update_rpc()
        try:
            logger.info(f"STARTING bitcoind {self.name} from status {self.status}")
            self.bitcoind.start_bitcoind(
                datadir=os.path.expanduser(self.datadir),
                timeout=timeout,  # At the initial startup, we don't wait on bitcoind
            )
            self._status = self.RUNNING
        except ExtProcTimeoutException as e:
            logger.error(e)
            e.check_logfile(os.path.join(self.datadir, "debug.log"))
            self._status = self.BROKEN
            logger.error(e.get_logger_friendly())
        except SpecterError as e:
            self._status = self.BROKEN
            logger.error(e)
            # Likely files of bitcoind were not found. Maybe deleted by the user?
        logger.info(f"STARTUP process complete {self.name} to status {self.status}")
        return self.update_rpc()

    def _update_status(self, bitcoin, status):
        self._bitcoin = bitcoin
        self._status = status

    def stop(self):
        logger.info(f"STOPPING bitcoind {self.name} from status {self.status}")
        success = self._bitcoind.stop_bitcoind()
        if success:
            self._update_status(None, self.DOWN)
            logger.info(f"bitcoind {self.name} stopped")
        else:
            logger.error(f"Failed to stop bitcoind {self.name}")
            self._status = self.BROKEN
        logger.info(f"STOPPING process complete {self.name} to status {self.status}")

    @property
    def bitcoind(self):
        if os.path.isfile(self.bitcoind_path):
            if not self._bitcoind:
                self._bitcoind = BitcoindPlainController(
                    bitcoind_path=self.bitcoind_path,
                    rpcport=self.port,
                    network=self.bitcoind_network,
                    rpcuser=self.user,
                    rpcpassword=self.password,
                )
            return self._bitcoind
        raise SpecterError(
            "Bitcoin Core files missing. Make sure Bitcoin Core is installed within Specter"
        )

    @property
    def status(self):
        """RUNNING DOWN or BROKEN."""
        return self._status

    def is_bitcoind_running(self):
        if self._status == self.RUNNING:
            return self._bitcoind and self._bitcoind.check_existing()
        return False

    @classmethod
    def kill_process(cls, proc_id):
        try:
            proc = psutil.Process(proc_id)
            proc.terminate()
            # os.kill(proc_id, signal.SIGTERM)
        except ProcessLookupError:
            logger.error("Process with ID {proc_id} does not exist")


class BitcoindProcess:
    """This is a class which represents a Bitcoind-process which is detected out of the context it has been created
    it wraps a Process and therefore behaves like one. At least until it doesn't
    """

    def __init__(self, pid):
        self.pid = pid
        self.proc = psutil.Process(self.pid)

    @classmethod
    def by_password(cls, password):
        """returns a BitcoindProcess which fits the rpcpassword == password"""
        pid = cls.get_process_id(password)
        if not pid is None:
            return BitcoindProcess(pid)
        return None

    def terminate(self):
        try:
            self.proc.terminate()
        except ProcessLookupError:
            logger.error("Process with ID {proc_id} does not exist")

    def kill(self):
        self.proc.kill()

    def poll(self):
        return self.proc.poll()

    def get_cmd_arg_value(self, arg_key):
        """parses the cmdline of the bitcoind-cmd and returns the value to the corresponding key
        e.g. for -rpcpassword=secret
        get_cmd_arg_value("rpcpassword") == "secret"
        """
        for cmd_arg in self.proc.cmdline():
            if cmd_arg.startswith(f"-{arg_key}"):
                arg_value = cmd_arg.split("=")[1]
                return arg_value

    def __repr__(self) -> str:
        return f"<BitcoindProcess datadir={self.get_cmd_arg_value('datadir')} rpcport={self.get_cmd_arg_value('rpcport')} >"

    @classmethod
    def get_process_id(cls, searched_password):
        for proc in psutil.process_iter():
            # logger.debug(f"investigating {proc.name()}")
            try:
                # Get process name & pid from process object.
                if not proc.name().endswith("bitcoind"):
                    continue
                if not proc.exe().endswith("/bitcoin-binaries/bin/bitcoind"):
                    continue
                password = None
                for cmd_arg in proc.cmdline():
                    if cmd_arg.startswith("-rpcpassword"):
                        password = cmd_arg.split("=")[1]
                if searched_password == password:
                    logger.debug(
                        f"Found internal_node bitcoind process with the password given: {proc.pid}"
                    )
                    return proc.pid
            except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
                # processes are volatile
                pass
        return None



import os
from ..specter_error import SpecterError


def download_file(specter, url, file_path, setup_status_name, setup_status_stage):
    response = specter.requests_session().get(url, stream=True)
    with open(file_path, "wb") as f:
        total_length = float(response.headers["content-length"])
        downloaded = 0.0
        old_progress = 0
        specter.update_setup_status(setup_status_name, setup_status_stage)
        for chunk in response.iter_content(chunk_size=4096):
            if specter.setup_status["stage"] == "start":
                raise SpecterError("File download canceled")
            downloaded += len(chunk)
            f.write(chunk)
            new_progress = int((downloaded / total_length) * 10000) / 100
            if new_progress > old_progress:
                if new_progress > 100:
                    new_progress = 100
                old_progress = new_progress
                specter.update_setup_download_progress(setup_status_name, new_progress)



import logging
from .util.shell import get_last_lines_from_file

logger = logging.getLogger(__name__)


class SpecterError(Exception):
    """A SpecterError contains meaningfull messages which can be passed directly to the user"""

    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super(SpecterError, self).__init__(message)
        # potentially, we could add own stuff now here:


class ExtProcTimeoutException(Exception):
    """A Exception which is thrown because an external process timed out
    use check_logfile to get some lines in loglines
    probably improvable for processes which don't have a logfile
    """

    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super(ExtProcTimeoutException, self).__init__(message)

    def check_logfile(self, logfile_location):
        try:
            self.logfile_location = logfile_location
            self.loglines = "     ".join(get_last_lines_from_file(logfile_location))
        except IOError as ioe:
            self.message = str(self) + f" ({ioe})"
            self.loglines = ""

    def get_logger_friendly(self):
        return f"""----------- {self.logfile_location} -----------
        {self.loglines}
        ------------------------------------------------------"""


def handle_exception(exception, user=None):
    """prints the exception and most important the stacktrace"""
    logger.error("Unexpected error:")
    logger.error(
        "----START-TRACEBACK-----------------------------------------------------------------"
    )
    logger.exception(exception)  # the exception instance
    logger.error(
        "----END---TRACEBACK-----------------------------------------------------------------"
    )



"""
The goal of this module is to slowly migrate from json-like representation of PSBT received from Bitcoin RPC
to a normal PSBT class that does not require RPC calls and can do more things.
to_dict and from_dict methods are maintained for backward-compatibility
"""
from cryptoadvance.specter.key import Key
from embit.psbt import PSBT, InputScope, OutputScope, DerivationPath
from embit.transaction import Transaction, TransactionOutput, TransactionInput
from embit.liquid.networks import get_network
from embit import bip32
from embit.descriptor import Descriptor
from math import ceil
import time
from typing import Union, Tuple, List


class AbstractTxContext:
    """Class inherited from this one must have the following properties:
    - self.network : dict with network constants (see embit.networks)
    - self.descriptor : Descriptor class that can check if it owns a PSBT scope or not
    Allows to pass context Wallet -> SpecterPSBT -> SpecterScope -> SpecterTx
    """

    @property
    def network(self) -> dict:
        if hasattr(self, "parent"):
            return self.parent.network
        raise NotImplementedError("Implement this!")

    @property
    def descriptor(self) -> Descriptor:
        if hasattr(self, "parent"):
            return self.parent.descriptor
        raise NotImplementedError("Implement this!")


class SpecterTx(AbstractTxContext):
    TxCls = Transaction

    def __init__(self, parent: AbstractTxContext, tx: Union[Transaction, str, bytes]):
        self.parent = parent
        if isinstance(tx, str):
            tx = self.TxCls.from_string(tx)
        elif isinstance(tx, bytes):
            tx = self.TxCls.parse(tx)
        self.tx = tx

    def vin_to_dict(self, vin: TransactionInput) -> dict:
        return {
            "txid": vin.txid.hex(),
            "vout": vin.vout,
            "sequence": vin.sequence,
        }

    def vout_to_dict(self, vout: TransactionOutput) -> dict:
        i = self.tx.vout.index(vout)
        obj = {
            "value": round(1e-8 * vout.value, 8),
            "sats": vout.value,
            "n": i,
            "scriptPubKey": {
                "hex": vout.script_pubkey.data.hex(),
            },
        }
        try:
            if scope.script_pubkey.data.startswith(b"\x6a"):
                obj["scriptPubKey"]["addresses"] = [
                    "OP_RETURN " + scope.script_pubkey.data.hex()
                ]
            else:
                obj["scriptPubKey"]["addresses"] = [
                    vout.script_pubkey.address(self.network)
                ]
        except:
            pass
        return obj

    def to_dict(self) -> dict:
        if not self.tx:
            return {}
        txid = self.tx.txid().hex()
        size = len(self.tx.serialize())
        return {
            "txid": txid,
            "hash": txid,  # not sure why it's the same
            "version": self.tx.version,
            "size": size,
            "vsize": size,
            "weight": 4 * size,
            "locktime": self.tx.locktime,
            "vin": [self.vin_to_dict(vin) for vin in self.tx.vin],
            "vout": [self.vout_to_dict(vout) for vout in self.tx.vout],
        }


class SpecterScope(AbstractTxContext):
    def __init__(
        self, parent: AbstractTxContext, scope: Union[InputScope, OutputScope]
    ):
        self.parent = parent
        self.scope = scope

    @property
    def is_mine(self) -> bool:
        return self.descriptor.owns(self.scope)

    @property
    def is_change(self) -> bool:
        """Returns True only if the scope belongs to change descriptor (branch 1)"""
        return self.descriptor.branch(1).owns(self.scope)

    @property
    def is_receiving(self) -> bool:
        return self.is_mine and not self.is_change

    @property
    def address(self) -> str:
        try:
            if self.scope.script_pubkey.data.startswith(b"\x6a"):
                return "OP_RETURN " + self.scope.script_pubkey.data.hex()
            else:
                return self.scope.script_pubkey.address(self.network)
        except:
            return None

    @property
    def sat_amount(self) -> int:
        """Implement this!"""
        raise NotImplementedError("Not implemented for this scope")

    @property
    def float_amount(self) -> float:
        return round(self.sat_amount * 1e-8, 8)

    def to_dict(self) -> dict:
        addr = self.address
        try:
            sats = self.sat_amount
        except:
            sats = None
        obj = {
            "change": self.is_change,
            "is_mine": self.is_mine,
        }
        if addr:
            obj["address"] = addr
        if sats is not None:
            obj.update(
                {
                    "float_amount": round(sats * 1e-8, 8),
                    "sat_amount": sats,
                }
            )
        if self.scope.bip32_derivations:
            obj["bip32_derivs"] = [
                {
                    "pubkey": pub.sec().hex(),
                    "master_fingerprint": der.fingerprint.hex(),
                    "path": bip32.path_to_str(der.derivation),
                }
                for pub, der in self.scope.bip32_derivations.items()
            ]
        return obj


class SpecterInputScope(SpecterScope):
    TxCls = SpecterTx

    @property
    def inp(self) -> InputScope:
        return self.scope

    @property
    def sat_amount(self) -> int:
        return self.scope.utxo.value

    @property
    def txid(self) -> bytes:
        return self.scope.txid

    @property
    def vout(self) -> int:
        return self.scope.vout

    def to_dict(self) -> dict:
        obj = super().to_dict()
        obj.update(
            {
                "txid": self.scope.txid.hex(),
                "vout": self.scope.vout,
            }
        )
        if self.scope.witness_utxo:
            obj["witness_utxo"] = {
                "amount": self.float_amount,
                "sats": self.sat_amount,
                "scriptPubKey": {
                    "hex": self.scope.script_pubkey.data.hex(),
                    "addresses": [self.address],
                },
            }
        else:
            obj["non_witness_utxo"] = self.TxCls(
                self, self.scope.non_witness_utxo
            ).to_dict()
        return obj


class SpecterOutputScope(SpecterScope):
    @property
    def out(self) -> OutputScope:
        return self.scope

    @property
    def sat_amount(self) -> int:
        return self.out.value


class SpecterPSBT(AbstractTxContext):
    """Specter's PSBT class with some handy functions"""

    PSBTCls = PSBT
    InputCls = SpecterInputScope
    OutputCls = SpecterOutputScope
    TxCls = SpecterTx

    def __init__(
        self,
        psbt: Union[str, PSBT],
        descriptor: Descriptor,
        network: dict,
        raw: Union[None, str] = None,
        devices: List[Tuple[Key, str]] = [],  # list of tuples: (Key, device_alias)
        **kwargs
    ):
        """
        kwargs can contain:
        - "time" - creation time of the transaction, time.time() is used if missing,
        - other keys in kwargs are dropped
        """
        if isinstance(psbt, str):
            psbt = self.PSBTCls.from_string(psbt)
        self.psbt = psbt
        self._descriptor = descriptor
        self._network = network
        self.devices = devices
        self.raw = bytes.fromhex(raw) if raw else None
        self.time = kwargs.get("time", time.time())

    @property
    def network(self) -> dict:
        return self._network

    @property
    def descriptor(self) -> Descriptor:
        return self._descriptor

    def update(self, b64psbt: str, raw: dict = {}) -> None:
        """
        b64psbt - PSBT transaction with some extra data that we should take
        raw dict can contain "hex" key with hex string of the finalized transaction. Or not.
        """
        if raw and "hex" in raw:
            self.raw = bytes.fromhex(raw["hex"])
        if not b64psbt:
            return
        psbt = self.PSBTCls.from_string(b64psbt)
        for inp1, inp2 in zip(self.psbt.inputs, psbt.inputs):
            inp1.update(inp2)
        for out1, out2 in zip(self.psbt.outputs, psbt.outputs):
            out1.update(out2)

    def utxo_dict(self) -> dict:
        return [{"txid": inp.txid.hex(), "vout": inp.vout} for inp in self.psbt.inputs]

    @property
    def extra_input_weight(self) -> int:
        redeem_script = self.descriptor.redeem_script()
        witness_script = self.descriptor.witness_script()
        weight = 0
        if redeem_script:
            weight += len(redeem_script.data) * 4
        if witness_script:
            weight += (
                len(witness_script.data) + 2
            )  # number of items in witness + script length
            if self.descriptor.is_basic_multisig:
                threshold = self.descriptor.miniscript.args[0].num
                num_keys = len(self.descriptor.keys)
                weight += num_keys * 34
                weight += threshold * 75
        else:
            # pubkey, signature
            weight += 75 + 34
        return weight

    @property
    def full_size(self) -> int:
        weight = len(self.psbt.tx.serialize()) * 4 + 4  # marker will be added
        weight += len(self.inputs) * self.extra_input_weight
        return ceil(weight / 4)

    @property
    def fee(self) -> int:
        return self.psbt.fee()

    @property
    def fee_rate(self) -> float:
        return self.fee / self.full_size

    @property
    def threshold(self) -> int:
        if self.descriptor.is_basic_multisig:
            return self.descriptor.miniscript.args[0].num
        return 1

    @property
    def sigs_count(self) -> int:
        # everything is signed if final witness is there or
        if self.raw or any([inp.final_scriptwitness for inp in self.psbt.inputs]):
            return self.threshold
        # not quite true but ok for most common cases
        return max([len(inp.partial_sigs) for inp in self.psbt.inputs])

    def should_display(self, out: SpecterOutputScope) -> bool:
        """Checks if this output should be displayed"""
        return not self.descriptor.branch(1).owns(out.scope)

    @property
    def addresses(self) -> List[str]:
        return [out.address for out in self.outputs if self.should_display(out)]

    @property
    def amounts(self) -> List[float]:
        return [out.float_amount for out in self.outputs if self.should_display(out)]

    @property
    def sats(self) -> List[int]:
        return [out.value for out in self.psbt.outputs if not self.descriptor.owns(out)]

    @property
    def txid(self) -> str:
        return self.psbt.tx.txid().hex()

    @property
    def inputs(self) -> List[SpecterInputScope]:
        return [self.InputCls(self, inp) for inp in self.psbt.inputs]

    @property
    def outputs(self) -> List[SpecterOutputScope]:
        return [self.OutputCls(self, out) for out in self.psbt.outputs]

    @property
    def tx(self) -> Transaction:
        return self.TxCls(self, self.psbt.tx)

    def get_signed_devices(self) -> List[str]:
        if not self.devices:
            return []
        devices = []
        # for each devices check if there is a partial signature for any input
        for key, device in self.devices:
            device_signed = False
            for inp in self.psbt.inputs:
                if not inp.partial_sigs:
                    continue
                for pub in inp.partial_sigs:
                    der = inp.bip32_derivations.get(pub)
                    if der and der.fingerprint.hex() == key.fingerprint:
                        device_signed = True
                        break
                if device_signed:
                    break
            if device_signed and device not in devices:
                devices.append(device)
        return devices

    @classmethod
    def from_dict(cls, obj: dict, descriptor: Descriptor, network: dict, devices=[]):
        psbt = cls.PSBTCls.from_string(obj["base64"])
        kwargs = {}
        kwargs.update(obj)
        kwargs.pop("base64")
        return cls(psbt, descriptor, network, devices=devices, **kwargs)

    @classmethod
    def from_transaction(
        cls,
        tx: Union[Transaction, str, bytes],
        descriptor: Descriptor,
        network: dict,
        devices=[],
    ):
        if isinstance(tx, str):
            tx = cls.TxCls.from_string(tx)
        elif isinstance(tx, bytes):
            tx = cls.TxCls.parse(tx)
        psbt = cls.PSBTCls(tx)
        return cls(psbt, descriptor, network, devices=devices)

    def to_dict(self) -> dict:
        # fee calculation may fail if inputs info is missing
        try:
            fee = self.fee
        except:
            fee = 0
        full_size = self.full_size
        obj = {
            "tx": self.tx.to_dict(),
            "inputs": [inp.to_dict() for inp in self.inputs],
            "outputs": [out.to_dict() for out in self.outputs],
            "base64": str(self.psbt),
            "fee": round(fee * 1e-8, 8),
            "fee_sat": fee,
            "address": self.addresses,
            "amount": self.amounts,
            "sats": self.sats,
            "tx_full_size": full_size,
            "sigs_count": self.sigs_count,
            "time": self.time,
            "devices_signed": [
                device if isinstance(device, str) else device.alias
                for device in self.get_signed_devices()
            ],
            "fee_rate": round(fee / full_size, 2),
        }
        if self.raw:
            obj.update({"raw": self.raw.hex()})
        return obj

    @classmethod
    def from_string(cls, b64psbt: str) -> PSBT:
        """Returns PSBTCls, not cls"""
        return cls.PSBTCls.from_string(b64psbt)

    def to_string(self) -> str:
        return str(self.psbt)

    def __str__(self):
        return str(self.psbt)

    @classmethod
    def fill_output(cls, out: OutputScope, desc: Descriptor) -> bool:
        """
        Fills derivations and all other information in PSBT output
        from derived descriptor
        """
        if desc.script_pubkey() != out.script_pubkey:
            return False
        out.redeem_script = desc.redeem_script()
        out.witness_script = desc.witness_script()
        out.bip32_derivations = {
            key.get_public_key(): DerivationPath(
                key.origin.fingerprint, key.origin.derivation
            )
            for key in desc.keys
        }
        return True



import logging
import re
from embit import bip32, ec, networks, script
from embit.liquid.networks import get_network
from cryptoadvance.specter.key import Key
from cryptoadvance.specter.specter_error import SpecterError

# Based on hwilib by achow101: https://github.com/bitcoin-core/HWI/blob/1.2.1/hwilib/descriptor.py which is from
# https://github.com/bitcoin/bitcoin/blob/0.21/src/script/descriptor.cpp


def PolyMod(c, val):
    c0 = c >> 35
    c = ((c & 0x7FFFFFFFF) << 5) ^ val
    if c0 & 1:
        c ^= 0xF5DEE51989
    if c0 & 2:
        c ^= 0xA9FDCA3312
    if c0 & 4:
        c ^= 0x1BAB10E32D
    if c0 & 8:
        c ^= 0x3706B1677A
    if c0 & 16:
        c ^= 0x644D626FFD
    return c


def DescriptorChecksum(desc):
    INPUT_CHARSET = "0123456789()[],'/*abcdefgh@:$%{}IJKLMNOPQRSTUVWXYZ&+-.;<=>?!^_|~ijklmnopqrstuvwxyzABCDEFGH`#\"\\ "
    CHECKSUM_CHARSET = "qpzry9x8gf2tvdw0s3jn54khce6mua7l"

    c = 1
    cls = 0
    clscount = 0
    for ch in desc:
        pos = INPUT_CHARSET.find(ch)
        if pos == -1:
            return ""
        c = PolyMod(c, pos & 31)
        cls = cls * 3 + (pos >> 5)
        clscount += 1
        if clscount == 3:
            c = PolyMod(c, cls)
            cls = 0
            clscount = 0
    if clscount > 0:
        c = PolyMod(c, cls)
    for j in range(0, 8):
        c = PolyMod(c, 0)
    c ^= 1

    ret = [None] * 8
    for j in range(0, 8):
        ret[j] = CHECKSUM_CHARSET[(c >> (5 * (7 - j))) & 31]
    return "".join(ret)


def AddChecksum(desc):
    return desc + "#" + DescriptorChecksum(desc)


# Not in hwilib
def derive_pubkey(key, path_suffix=None, idx=None):
    # if SEC pubkey - just return it
    if key[:2] in ["02", "03", "04"]:
        return ec.PublicKey.parse(bytes.fromhex(key))
    # otherwise - xpub or xprv
    hd = bip32.HDKey.from_base58(key)
    if hd.is_private:
        hd = hd.to_public()
    # if we have path suffix
    path = "m" + (path_suffix or "")
    if idx is not None:
        path = path.replace("*", str(idx))
    return hd.derive(path).key


class Descriptor:
    def __init__(
        self,
        origin_fingerprint,
        origin_path,
        base_key,
        path_suffix,
        testnet,
        sh_wpkh=None,
        wpkh=None,
        sh=None,
        sh_wsh=None,
        wsh=None,
        multisig_M=None,
        multisig_N=None,
        sort_keys=True,
    ):
        self.origin_fingerprint = origin_fingerprint
        self.origin_path = origin_path
        self.path_suffix = path_suffix
        self.base_key = base_key
        self.testnet = testnet
        self.sh_wpkh = sh_wpkh
        self.wpkh = wpkh
        self.sh = sh
        self.sh_wsh = sh_wsh
        self.wsh = wsh
        self.multisig_M = multisig_M
        self.multisig_N = multisig_N
        self.m_path = None
        self.sort_keys = sort_keys

        if origin_path and not isinstance(origin_path, list):
            self.m_path_base = "m" + origin_path
            self.m_path = "m" + origin_path + (path_suffix or "")
        elif isinstance(origin_path, list):
            self.m_path_base = []
            self.m_path = []
            for i in range(0, len(origin_path)):
                if origin_path[i]:
                    self.m_path_base.append("m" + origin_path[i])
                    self.m_path.append("m" + origin_path[i] + (path_suffix[i] or ""))
                else:
                    self.m_path_base.append(None)
                    self.m_path.append(None)

    @classmethod
    def parse(cls, desc, testnet=False):
        sh_wpkh = None
        wpkh = None
        sh = None
        sh_wsh = None
        wsh = None
        origin_fingerprint = None
        origin_path = None
        base_key_and_path_match = None
        base_key = None
        path_suffix = None
        multisig_M = None
        multisig_N = None
        sort_keys = True

        # Check the checksum
        check_split = desc.split("#")
        # Multiple # in desc
        if len(check_split) > 2:
            raise SpecterError(
                f"Too many separators in the descriptor. Check if there are multiple # in {desc}."
            )
        if len(check_split) == 2:
            # Empty checkusm
            if len(check_split[1]) == 0:
                raise SpecterError("Checksum is empty.")
            # Incorrect length
            elif len(check_split[1]) != 8:
                raise SpecterError(
                    f"Checksum {check_split[1]} doesn't have the correct length. Should be 8 characters not {len(check_split[1])}."
                )
            checksum = DescriptorChecksum(check_split[0])
            # Check of checksum calc
            if checksum.strip() == "":
                raise SpecterError(f"Checksum calculation went wrong.")
            # Wrong checksum
            if checksum != check_split[1]:
                raise SpecterError(
                    f"{check_split[1]} is the wrong checkum should be {checksum}."
                )
        desc = check_split[0]

        if desc.startswith("sh(wpkh("):
            sh_wpkh = True
        elif desc.startswith("wpkh("):
            wpkh = True
        elif desc.startswith("sh(wsh("):
            sh_wsh = True
        elif desc.startswith("wsh("):
            wsh = True
        elif desc.startswith("sh("):
            sh = True

        if sh or sh_wsh or wsh:
            if "multi(" not in desc:
                # only multisig scripts are supported
                return None
            # get the list of keys only
            keys = desc.split(",", 1)[1].split(")", 1)[0].split(",")
            sort_keys = "sortedmulti" in desc
            if "sortedmulti" in desc:
                # sorting makes sense only if individual pubkeys are provided
                base_keys = [x if "]" not in x else x.split("]")[1] for x in keys]
                bare_pubkeys = [k for k in base_keys if k[:2] in ["02", "03", "04"]]
                if len(bare_pubkeys) == len(keys):
                    keys.sort(key=lambda x: x if "]" not in x else x.split("]")[1])
            multisig_M = desc.split(",")[0].split("(")[-1]
            multisig_N = len(keys)
            if int(multisig_M) > multisig_N:
                raise SpecterError(
                    f"Multisig threshold cannot be larger than the number of keys. Threshold is {int(multisig_M)} but only {multisig_N} keys specified."
                )
        else:
            keys = [desc.split("(")[-1].split(")", 1)[0]]

        descriptors = []
        for key in keys:
            origin_fingerprint = None
            origin_path = None
            base_key = None
            path_suffix = None
            origin_match = re.search(r"\[(.*)\]", key)
            if origin_match:
                origin = origin_match.group(1)
                match = re.search(r"^([0-9a-fA-F]{8})(\/.*)", origin)
                if match:
                    origin_fingerprint = match.group(1)
                    origin_path = match.group(2)
                    # Replace h with '
                    origin_path = origin_path.replace("h", "'")
                else:
                    origin_fingerprint = origin
                    origin_path = ""

                base_key_and_path_match = re.search(r"\[.*\](\w+)([\d'\/\*]*)", key)
            else:
                base_key_and_path_match = re.search(r"(\w+)([\d'\/\*]*)", key)

            if base_key_and_path_match:
                base_key = base_key_and_path_match.group(1)
                path_suffix = base_key_and_path_match.group(2)
                if path_suffix == "":
                    path_suffix = None
            else:
                if origin_match is None:
                    return None

            descriptors.append(
                cls(
                    origin_fingerprint,
                    origin_path,
                    base_key,
                    path_suffix,
                    testnet,
                    sh_wpkh,
                    wpkh,
                    sh,
                    sh_wsh,
                    wsh,
                    sort_keys,
                )
            )

        if len(descriptors) == 1:
            return descriptors[0]
        else:
            # for multisig scripts save as lists all keypaths fields
            return cls(
                [descriptor.origin_fingerprint for descriptor in descriptors],
                [descriptor.origin_path for descriptor in descriptors],
                [descriptor.base_key for descriptor in descriptors],
                [descriptor.path_suffix for descriptor in descriptors],
                testnet,
                sh_wpkh,
                wpkh,
                sh,
                sh_wsh,
                wsh,
                multisig_M,
                multisig_N,
                sort_keys,
            )

    @property
    def is_multisig(self):
        return bool(self.multisig_N)

    @property
    def address_type(self):
        if self.wpkh:
            return "wpkh"
        elif self.wsh:
            return "wsh"
        elif self.sh_wpkh:
            return "sh-wpkh"
        elif self.sh_wsh:
            return "sh-wsh"
        elif self.sh:
            return "sh"
        else:
            return "pkh"

    def derive(self, idx, keep_xpubs=False):
        """
        Derives a descriptor with index idx up to the pubkeys.
        If keep_xpubs is False all xpubs will be replaced by pubkeys
        so [fgp/path]xpub/suffix changes to [fgp/path/suffix]pubkey
        Otherwise xpubs will be sorted according to pubkeys
        but remain in the descriptor
        """
        if self.is_multisig:
            keys = []
            for i, key in enumerate(self.base_key):
                keys.append(
                    (
                        derive_pubkey(key, self.path_suffix[i], idx),
                        key,
                        self.origin_fingerprint[i],
                        self.origin_path[i],
                        self.path_suffix[i],
                    )
                )
            if self.sort_keys:
                keys = sorted(keys, key=lambda k: k[0])
            origin_fingerprint = [k[2] for k in keys]
            if keep_xpubs:
                base_key = [k[1] for k in keys]
                origin_path = [k[3] for k in keys]
                path_suffix = [(k[4] or "") for k in keys]
                path_suffix = [(p.replace("*", str(idx)) or None) for p in path_suffix]
            else:
                base_key = [k[0].sec().hex() for k in keys]
                origin_path = [(k[3] or "") + (k[4] or "") for k in keys]
                origin_path = [(p.replace("*", str(idx)) or None) for p in origin_path]
                path_suffix = [None for k in keys]
        else:
            origin_fingerprint = self.origin_fingerprint
            if keep_xpubs:
                base_key = self.base_key
                origin_path = self.origin_path
                path_suffix = self.path_suffix
                path_suffix = (
                    path_suffix.replace("*", str(idx)) if path_suffix else None
                )
            else:
                base_key = (
                    derive_pubkey(self.base_key, self.path_suffix, idx).sec().hex()
                )
                origin_path = (self.origin_path or "") + (self.path_suffix or "")
                origin_path = origin_path.replace("*", str(idx)) or None
                path_suffix = None
        return Descriptor(
            origin_fingerprint,
            origin_path,
            base_key,
            path_suffix,
            self.testnet,
            self.sh_wpkh,
            self.wpkh,
            self.sh,
            self.sh_wsh,
            self.wsh,
            self.multisig_M,
            self.multisig_N,
            self.sort_keys,
        )

    def scriptpubkey(self, idx=None):
        if idx is None and "*" in self.serialize():
            raise RuntimeError("Index is required")
        if self.is_multisig:
            keys = []
            for i, key in enumerate(self.base_key):
                keys.append(derive_pubkey(key, self.path_suffix[i], idx))
            if self.sort_keys:
                keys = sorted(keys)
            sc = script.multisig(int(self.multisig_M), keys)
            if self.sh:
                return script.p2sh(sc)
            elif self.sh_wsh:
                return script.p2sh(script.p2wsh(sc))
            elif self.wsh:
                return script.p2wsh(sc)
        else:
            key = derive_pubkey(self.base_key, self.path_suffix, idx)
            if self.wpkh:
                return script.p2wpkh(key)
            elif self.sh_wpkh:
                return script.p2sh(script.p2wpkh(key))
            else:
                return script.p2pkh(key)

    def address(self, idx=None, network=None):
        if network is None:
            net = get_network("test" if self.testnet else "main")
        else:
            net = get_network(network)
        return self.scriptpubkey(idx).address(net)

    def serialize(self):
        descriptor_open = "pkh("
        descriptor_close = ")"

        if self.wpkh:
            descriptor_open = "wpkh("
        elif self.sh_wpkh:
            descriptor_open = "sh(wpkh("
            descriptor_close = "))"
        elif self.sh:
            descriptor_open = "sh("
            descriptor_close = ")"
        elif self.sh_wsh:
            descriptor_open = "sh(wsh("
            descriptor_close = "))"
        elif self.wsh:
            descriptor_open = "wsh("
            descriptor_close = ")"

        if self.is_multisig:
            multi = "sortedmulti" if self.sort_keys else "multi"
            base_open = f"{multi}({self.multisig_M},"
            base_close = ")"
            origins = []
            for i in range(self.multisig_N):
                path_suffix = ""
                origin = ""
                if self.origin_fingerprint[i] and self.origin_path[i]:
                    origin = (
                        "[" + self.origin_fingerprint[i] + self.origin_path[i] + "]"
                    )

                if self.path_suffix[i]:
                    path_suffix = self.path_suffix[i]

                origins.append(origin + self.base_key[i] + path_suffix)

            base = base_open + ",".join(origins) + base_close
        else:
            origin = ""
            path_suffix = ""
            if self.origin_fingerprint and self.origin_path:
                origin = "[" + self.origin_fingerprint + self.origin_path + "]"

            if self.path_suffix:
                path_suffix = self.path_suffix

            base = origin + self.base_key + path_suffix

        return AddChecksum(descriptor_open + base + descriptor_close)

    def parse_signers(self, devices, cosigners_types):
        keys = []
        cosigners = []
        unknown_cosigners = []
        unknown_cosigners_types = []

        if self.multisig_N == None:
            self.multisig_N = 1
            self.multisig_M = 1
            self.origin_fingerprint = [self.origin_fingerprint]
            self.origin_path = [self.origin_path]
            self.base_key = [self.base_key]
        for i in range(self.multisig_N):
            cosigner_found = False
            for device in devices:
                cosigner = devices[device]
                if self.origin_fingerprint[i] is None:
                    self.origin_fingerprint[i] = ""
                if self.origin_path[i] is None:
                    self.origin_path[i] = self.origin_fingerprint[i]
                for key in cosigner.keys:
                    if key.fingerprint + key.derivation.replace(
                        "m", ""
                    ) == self.origin_fingerprint[i] + self.origin_path[i].replace(
                        "'", "h"
                    ):
                        keys.append(key)
                        cosigners.append(cosigner)
                        cosigner_found = True
                        break
                if cosigner_found:
                    break
            if not cosigner_found:
                desc_key = Key.parse_xpub(
                    "[{}{}]{}".format(
                        self.origin_fingerprint[i],
                        self.origin_path[i],
                        self.base_key[i],
                    )
                )
                if len(cosigners_types) > i:
                    unknown_cosigners.append((desc_key, cosigners_types[i]["label"]))
                else:
                    unknown_cosigners.append((desc_key, None))
                if len(unknown_cosigners) > len(cosigners_types):
                    unknown_cosigners_types.append("other")
                else:
                    unknown_cosigners_types.append(cosigners_types[i]["type"])

        return (keys, cosigners, unknown_cosigners, unknown_cosigners_types)


def sort_descriptor(descriptor, index=None):
    """
    Sorts descriptor to maintain compatibility with Core 19
    as it doesn't support sortedmulti.
    Returns a derived multi() descriptor with sorted xpubs inside.
    """
    desc = Descriptor.parse(descriptor)
    desc.sort_keys = True
    sorted_desc = desc.derive(index, keep_xpubs=True)
    sorted_desc.sort_keys = False
    return sorted_desc.serialize()



import logging
import os
import pathlib
import shutil
import threading
import traceback
from typing import Dict

from flask_babel import lazy_gettext as _
from cryptoadvance.specter.rpc import BitcoinRPC

from cryptoadvance.specter.key import Key

from ..helpers import add_dicts, alias, is_liquid, load_jsons
from ..liquid.wallet import LWallet
from ..persistence import delete_folder
from ..rpc import RpcError, get_default_datadir
from ..specter_error import SpecterError, handle_exception
from ..wallet import (  # TODO: `purposes` unused here, but other files rely on this import
    Wallet,
    purposes,
)

logger = logging.getLogger(__name__)


class WalletManager:
    """Manages Wallets. Depending on the chain"""

    # chain is required to manage wallets when bitcoind is not running
    def __init__(
        self,
        bitcoin_core_version_raw,
        data_folder,
        rpc,
        chain,
        device_manager,
        path="specter",
        allow_threading=True,
    ):
        self.data_folder = data_folder
        self.chain = chain
        self.rpcs = {}
        self.rpc = rpc
        self.rpc_path = path
        self.device_manager = device_manager
        # sort of lock to prevent threads to update in parallel
        self.is_loading = False
        # key is the name of the wallet, value is the actual instance

        self.wallets = {}
        # A way to communicate failed wallets to the outside
        self.bitcoin_core_version_raw = bitcoin_core_version_raw
        self.allow_threading = allow_threading
        # define different wallet classes for liquid and bitcoin
        self.WalletClass = LWallet if is_liquid(chain) else Wallet
        self.update(data_folder, rpc, chain)

    def update(
        self,
        data_folder: str = None,
        rpc: BitcoinRPC = None,
        chain: str = None,
        use_threading=True,
    ):
        """Restructures the instance, specifically if chain/rpc changed
        The _update internal method will resync the internal status with Bitcoin Core
        use_threading : for the _update method which is heavily communicating with Bitcoin Core
        """
        if (chain is None and rpc is not None) or (chain is not None and rpc is None):
            raise Exception(
                f"Chain ({chain}) and rpc ({rpc}) can only be changed with one another"
            )
        if self.is_loading:
            return
        self.is_loading = True
        if chain is not None:
            self.chain = chain
        if data_folder is not None:
            self.data_folder = data_folder
            if data_folder.startswith("~"):
                data_folder = os.path.expanduser(data_folder)
            # creating folders if they don't exist
            if not os.path.isdir(data_folder):
                os.makedirs(data_folder, exist_ok=True)
        if rpc is not None and rpc.test_connection():
            self.rpc = rpc
        else:
            if rpc:
                logger.error(
                    f"Prevented Trying to update wallet_Manager with broken {rpc}"
                )
        # wallets_update_list is something like:
        # {'Specter': {'name': 'Specter', 'alias': 'pacman', ... }, 'another_wallet': { ... } }
        # It contains the same data as the JSON on disk
        wallets_update_list = {}
        if self.working_folder is not None and self.rpc is not None:
            wallets_files = load_jsons(self.working_folder, key="name")
            for wallet in wallets_files:
                wallet_name = wallets_files[wallet]["name"]
                wallets_update_list[wallet_name] = wallets_files[wallet]
                wallets_update_list[wallet_name]["is_multisig"] = (
                    len(wallets_files[wallet]["keys"]) > 1
                )
                wallets_update_list[wallet_name]["keys_count"] = len(
                    wallets_files[wallet]["keys"]
                )
            if self.allow_threading and use_threading:
                t = threading.Thread(
                    target=self._update,
                    args=(wallets_update_list,),
                )

                t.start()
            else:
                self._update(wallets_update_list)
        else:
            self.is_loading = False
            logger.info(
                "Specter seems to be disconnected from Bitcoin Core. Skipping wallets update."
            )

    def _update(self, wallets_update_list: Dict):
        """Effectively a three way sync. The three data sources are:
        * the json on disk (wallets_update_list)
        * the current wallet instances (existing_names)
        * the loaded wallets from Bitcoin Core (loaded_wallets)
        So, if we have data on disk
        * we get a list of loaded wallets from Bitcoin Core
        * the unloaded wallets are loaded in Bitcoin Core
        * and, on the Specter side, the wallet objects of those unloaded wallets are reinitialised
        """
        # list of wallets in the dict
        existing_names = list(self.wallets.keys())
        # list of wallet to keep
        self._failed_load_wallets = []
        try:
            if wallets_update_list:
                loaded_wallets = self.rpc.listwallets()
                # logger.info("Getting loaded wallets list from Bitcoin Core")
                for wallet in wallets_update_list:
                    wallet_alias = wallets_update_list[wallet]["alias"]
                    wallet_name = wallets_update_list[wallet]["name"]
                    # wallet from json not yet loaded in Bitcoin Core?!
                    if os.path.join(self.rpc_path, wallet_alias) not in loaded_wallets:
                        try:
                            logger.info(
                                "Loading %s to Bitcoin Core"
                                % wallets_update_list[wallet]["alias"]
                            )
                            self.rpc.loadwallet(
                                os.path.join(self.rpc_path, wallet_alias)
                            )
                            logger.info(
                                "Initializing %s Wallet object"
                                % wallets_update_list[wallet]["alias"]
                            )
                            loaded_wallet = self.WalletClass.from_json(
                                wallets_update_list[wallet],
                                self.device_manager,
                                self,
                            )
                            # Lock UTXO of pending PSBTs
                            logger.info(
                                "Re-locking UTXOs of wallet %s"
                                % wallets_update_list[wallet]["alias"]
                            )
                            if len(loaded_wallet.pending_psbts) > 0:
                                for psbt in loaded_wallet.pending_psbts:
                                    logger.info(
                                        f"lock {wallet_alias} {loaded_wallet.pending_psbts[psbt].utxo_dict()}"
                                    )
                                    loaded_wallet.rpc.lockunspent(
                                        False,
                                        loaded_wallet.pending_psbts[psbt].utxo_dict(),
                                    )
                            if len(loaded_wallet.frozen_utxo) > 0:
                                loaded_wallet.rpc.lockunspent(
                                    False,
                                    [
                                        {
                                            "txid": utxo.split(":")[0],
                                            "vout": int(utxo.split(":")[1]),
                                        }
                                        for utxo in loaded_wallet.frozen_utxo
                                    ],
                                )
                            self.wallets[wallet_name] = loaded_wallet
                            # logger.info(
                            #     "Finished loading wallet into Bitcoin Core and Specter: %s"
                            #     % wallets_update_list[wallet]["alias"]
                            # )
                        except RpcError as e:
                            logger.warning(
                                f"Couldn't load wallet {wallet_alias} into Bitcoin Core. Silently ignored! RPC error: {e}"
                            )
                            self._failed_load_wallets.append(
                                {
                                    **wallets_update_list[wallet],
                                    "loading_error": str(e).replace("'", ""),
                                }
                            )
                        except Exception as e:
                            logger.warning(
                                f"Couldn't load wallet {wallet_alias}. Silently ignored! Wallet error: {e}"
                            )
                            self._failed_load_wallets.append(
                                {
                                    **wallets_update_list[wallet],
                                    "loading_error": str(e).replace("'", ""),
                                }
                            )
                    else:
                        if wallet_name not in existing_names:
                            # ok wallet is not yet in the dict, create one
                            try:
                                loaded_wallet = self.WalletClass.from_json(
                                    wallets_update_list[wallet],
                                    self.device_manager,
                                    self,
                                )
                                self.wallets[wallet_name] = loaded_wallet
                            except Exception as e:
                                handle_exception(e)
                                self._failed_load_wallets.append(
                                    {
                                        **wallets_update_list[wallet],
                                        "loading_error": str(e).replace("'", ""),
                                    }
                                )
                        else:
                            # Wallet is already there
                            # we only need to update
                            self.wallets[wallet_name].update()
        # only ignore rpc errors
        except RpcError as e:
            logger.error(f"Failed updating wallet manager. RPC error: {e}")
        # logger.info("Done updating wallet manager")
        wallets_update_list = {}
        self.is_loading = False

    def get_by_alias(self, alias):
        for wallet_name in self.wallets:
            if self.wallets[wallet_name] and self.wallets[wallet_name].alias == alias:
                return self.wallets[wallet_name]
        raise SpecterError("Wallet %s does not exist!" % alias)

    @property
    def failed_load_wallets(self) -> list:
        """A list of wallets failed where not loaded"""
        if not hasattr(self, "_failed_load_wallets"):
            self._failed_load_wallets = []
        return self._failed_load_wallets

    @property
    def working_folder(self):
        if self.data_folder is None or self.chain is None:
            return None
        working_folder = os.path.join(self.data_folder, self.chain)
        pathlib.Path(working_folder).mkdir(parents=True, exist_ok=True)
        return working_folder

    @property
    def wallets_names(self):
        return sorted(self.wallets.keys())

    @property
    def rpc(self):
        """returns a BitcoinRpc depending on the chain"""
        return self.rpcs[self.chain]

    @rpc.setter
    def rpc(self, value):
        """sets a BitcoinRpc depending on the chain. This is an internal property. Don't use it from outside. ToDo: Refactor"""
        self.rpcs[self.chain] = value

    @property
    def wallets(self) -> Dict[str, Dict]:
        """returns wallets depending on the chain"""
        if not hasattr(self, "_wallets"):
            self._wallets = {}
        if not self._wallets.get(self.chain):
            self._wallets[self.chain] = {}
        # _wallets should look like something:
        # { "regtest" : {
        #       "wallet_name_1": {...},
        #       "wallet_name_2": {...},
        #    },
        #   "main": ...
        # }}
        return self._wallets[self.chain]

    @wallets.setter
    def wallets(self, value):
        if not hasattr(self, "_wallets"):
            self._wallets = {}
        if not self._wallets.get(self.chain):
            self._wallets[self.chain] = {}
        self._wallets[self.chain] = value

    def create_wallet(self, name, sigs_required, key_type, keys, devices, **kwargs):
        try:
            walletsindir = [
                wallet["name"] for wallet in self.rpc.listwalletdir()["wallets"]
            ]
        except:
            walletsindir = []
        self._check_duplicate_keys(keys)
        wallet_alias = alias(name)
        i = 2
        while (
            os.path.isfile(os.path.join(self.working_folder, "%s.json" % wallet_alias))
            or os.path.join(self.rpc_path, wallet_alias) in walletsindir
        ):
            wallet_alias = alias("%s %d" % (name, i))
            i += 1

        w = self.WalletClass.create(
            self.rpc,
            self.rpc_path,
            self.working_folder,
            self.device_manager,
            self,
            name,
            wallet_alias,
            sigs_required,
            key_type,
            keys,
            devices,
            self.bitcoin_core_version_raw,
            **kwargs,
        )
        # save wallet file to disk
        w.save_to_file()
        # get Wallet class instance
        if w:
            self.wallets[name] = w
            logger.info(f"Successfully created Wallet {name}")
            return w
        else:
            raise ("Failed to create new wallet")

    def delete_wallet(
        self, wallet, bitcoin_datadir=get_default_datadir(), chain="main"
    ):
        logger.info("Deleting {}".format(wallet.alias))
        wallet_rpc_path = os.path.join(self.rpc_path, wallet.alias)
        self.rpc.unloadwallet(wallet_rpc_path)
        # Try deleting wallet folder
        if bitcoin_datadir:
            if chain != "main":
                bitcoin_datadir = os.path.join(bitcoin_datadir, chain)
            candidates = [
                os.path.join(bitcoin_datadir, wallet_rpc_path),
                os.path.join(bitcoin_datadir, "wallets", wallet_rpc_path),
            ]
            for path in candidates:
                shutil.rmtree(path, ignore_errors=True)

        # Delete files
        wallet.delete_files()
        del self.wallets[wallet.name]
        self.update()

    def rename_wallet(self, wallet, name):
        logger.info("Renaming {}".format(wallet.alias))
        if wallet.name not in self.wallets_names:
            raise Exception(
                f"Wallet {wallet.name} is not managed by this WalletManager or the wallet is on a different chain than {self.chain}"
            )
        self.wallets.pop(wallet.name)
        wallet.name = name
        if self.working_folder is not None:
            wallet.save_to_file()
        self.wallets[name] = wallet

    def joined_balance(self):
        """
        Joined balance of all wallets.
        I don't call it full balance because
        full balance is different - see wallet.fullbalance...
        """
        balance = {}
        for wallet in self.wallets.values():
            add_dicts(balance, wallet.balance)
        return balance

    def full_txlist(
        self,
        fetch_transactions=True,
        validate_merkle_proofs=False,
        current_blockheight=None,
        service_id=None,
    ):
        """Returns a list of all transactions in all wallets loaded in the wallet_manager.
        #Parameters:
        #    fetch_transactions (bool): Update the TxList CSV caching by fetching transactions from the Bitcoin RPC
        #    validate_merkle_proofs (bool): Return transactions with validated_blockhash
        #    current_blockheight (int): Current blockheight for calculating confirmations number (None will fetch the block count from the RPC)
        #    service_id (str): Filters results for just the specified Service
        """
        # Nested comprehensions:
        txlists = [
            [
                # Inner comprehension: Return each tx and all its attrs as a list of dicts...
                # TODO: Simplify this by adding an `as_dict` option to `Wallet.txlist()`?
                {**tx, "wallet_alias": wallet.alias}
                for tx in wallet.txlist(
                    fetch_transactions=fetch_transactions,
                    validate_merkle_proofs=validate_merkle_proofs,
                    current_blockheight=current_blockheight,
                    service_id=service_id,
                )
            ]
            # Outer comprehension: ...from each wallet, each returning their own tx list.
            for wallet in self.wallets.values()
        ]
        result = []
        for txlist in txlists:
            for tx in txlist:
                result.append(tx)
        return list(reversed(sorted(result, key=lambda tx: tx["time"])))

    def full_utxo(self):
        """Returns a list of all UTXOs in all wallets loaded in the wallet_manager."""
        txlists = [
            [
                {
                    **utxo,
                    "label": wallet.getlabel(utxo["address"]),
                    "wallet_alias": wallet.alias,
                }
                for utxo in wallet.full_utxo
            ]
            for wallet in self.wallets.values()
        ]
        result = []
        for txlist in txlists:
            for tx in txlist:
                result.append(tx)
        return list(reversed(sorted(result, key=lambda tx: tx["time"])))

    def full_addresses_info(self, is_change: bool = False, service_id: str = None):
        """Mimics full_txlist in concept, but is really only expected to be used for
        retrieving all addresses across all Wallets that are associated with a
        Service.

        Not currently used yet."""
        addresses_info = []
        for wallet_alias, wallet in self.wallets.items():
            addresses_info.extend(
                wallet.addresses_info(
                    is_change=is_change,
                    service_id=service_id,
                    include_wallet_alias=True,
                )
            )
        return addresses_info

    def delete(self, specter):
        """Deletes all the wallets"""
        for w in list(self.wallets.keys()):
            wallet = self.wallets[w]
            self.delete_wallet(wallet, specter.bitcoin_datadir, specter.chain)
        delete_folder(self.data_folder)

    @classmethod
    def _check_duplicate_keys(cls, keys):
        """raise a SpecterError when a xpub in the passed KeyList is listed twice. Should prevent MultisigWallets where
        xpubs are used twice.
        """
        # normalizing xpubs in order to ignore slip132 differences
        xpubs = [Key.parse_xpub(key.original).xpub for key in keys]
        for xpub in xpubs:
            if xpubs.count(xpub) > 1:
                raise SpecterError(_(f"xpub {xpub} seem to be used at least twice!"))



import json
import logging
import random
from functools import wraps

import requests
from cryptoadvance.specter.commands.psbt_creator import PsbtCreator
from cryptoadvance.specter.util.wallet_importer import WalletImporter
from cryptoadvance.specter.wallet import Wallet
from cryptoadvance.specter.util.tx import is_hex, convert_rawtransaction_to_psbt
from flask import Blueprint
from flask import current_app as app
from flask import flash, jsonify, redirect, render_template, request, url_for
from flask_babel import lazy_gettext as _
from flask_login import login_required

from ..helpers import bcur2base64, get_devices_with_keys_by_type, get_txid
from ..key import Key
from ..managers.wallet_manager import purposes
from ..persistence import delete_file
from ..specter_error import SpecterError, handle_exception

logger = logging.getLogger(__name__)

rand = random.randint(0, 1e32)  # to force style refresh

# Setup endpoint blueprint
wallets_endpoint = Blueprint("wallets_endpoint", __name__)


def handle_wallet_error(func_name, error):
    flash(_("SpecterError while {}: {}").format(func_name, error), "error")
    app.logger.error(f"SpecterError while {func_name}: {error}")
    app.specter.wallet_manager.update()
    return redirect(url_for("welcome_endpoint.about"))


def check_wallet(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        """checks the wallet for healthiness A wrapper function"""
        if kwargs["wallet_alias"]:
            wallet_alias = kwargs["wallet_alias"]
            wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
            wallet.get_info()
        return func(*args, **kwargs)

    return wrapper


################## Wallet overview #######################


@wallets_endpoint.route("/wallets_overview/")
@login_required
def wallets_overview():
    app.specter.check_blockheight()
    for wallet in list(app.specter.wallet_manager.wallets.values()):
        wallet.update_balance()
        wallet.check_utxo()

    return render_template(
        "wallet/wallets_overview.jinja",
        specter=app.specter,
        rand=rand,
        services=app.specter.service_manager.services,
    )


################## Failed wallets fix #######################


@wallets_endpoint.route("/failed_wallets/", methods=["POST"])
@login_required
def failed_wallets():
    if request.method == "POST":
        action = request.form["action"]
        if action == "retry_loading_wallets":
            app.specter.wallet_manager.update()
        elif action == "delete_failed_wallet":
            try:
                wallet = json.loads(request.form["wallet_data"])
                fullpath = wallet["fullpath"]
                delete_file(fullpath)
                delete_file(fullpath + ".bkp")
                delete_file(fullpath.replace(".json", "_addr.csv"))
                delete_file(fullpath.replace(".json", "_txs.csv"))
                app.specter.wallet_manager.update()
            except Exception as e:
                handle_exception(e)
                flash(_("Failed to delete wallet: {}").format(str(e)), "error")
    return redirect("/")


################## New wallet #######################


@wallets_endpoint.route("/new_wallet/")
@login_required
def new_wallet_type():
    err = None
    if app.specter.chain is None:
        err = _("Configure Bitcoin Core to create wallets")
        return render_template("base.jinja", error=err, specter=app.specter, rand=rand)
    try:
        # Make sure wallet is enabled on Bitcoin Core
        app.specter.rpc.listwallets()
    except Exception as e:
        handle_exception(e)
        # Hmm, would be better to be more precise with this exception. Best assumption:
        err = _(
            '<p><br>Bitcoin Core is running with wallets disabled.<br><br>Please make sure disablewallet is off (set disablewallet=0 in your bitcoin.conf), then restart Bitcoin Core and try again.<br>See <a href="https://github.com/cryptoadvance/specter-desktop/blob/34ca139694ecafb2e7c2bd5ad5c4ac74c6d11501/docs/faq.md#im-not-sure-i-want-the-bitcoin-core-wallet-functionality-to-be-used-is-that-mandatory-if-so-is-it-considered-secure" target="_blank" style="color: white;">here</a> for more information.</p>'
        )
        return render_template("base.jinja", error=err, specter=app.specter, rand=rand)
    return render_template(
        "wallet/new_wallet/new_wallet_type.jinja", specter=app.specter, rand=rand
    )


@wallets_endpoint.route("/new_wallet/<wallet_type>/", methods=["GET", "POST"])
@login_required
def new_wallet(wallet_type):
    wallet_types = ["simple", "multisig", "import_wallet"]
    if wallet_type not in wallet_types:
        flash(_("Unknown wallet type requested"), "error")
        return redirect(url_for("wallets_endpoint.new_wallet_type"))

    err = None
    if request.method == "POST":
        action = request.form["action"]
        if action == "importwallet":
            try:
                wallet_importer: WalletImporter = WalletImporter(
                    request.form["wallet_data"],
                    app.specter,
                )
            except SpecterError as se:
                flash(str(se), "error")
                return redirect(url_for("wallets_endpoint.new_wallet_type"))
            createwallet = "createwallet" in request.form
            if createwallet:
                # User might have renamed it
                wallet_importer.wallet_name = request.form["wallet_name"]
                wallet_importer.create_nonexisting_signers(
                    app.specter.device_manager, request.form
                )
                try:
                    wallet_importer.create_wallet(app.specter.wallet_manager)
                except SpecterError as se:
                    flash(str(se), "error")
                    return redirect(url_for("wallets_endpoint.new_wallet_type"))
                flash(_("Wallet imported successfully"), "info")
                try:
                    wallet_importer.rescan_as_needed(app.specter)
                except SpecterError as se:
                    flash(str(se), "error")
                return redirect(
                    url_for(
                        "wallets_endpoint.receive",
                        wallet_alias=wallet_importer.wallet.alias,
                    )
                    + "?newwallet=true"
                )
            else:
                return render_template(
                    "wallet/new_wallet/import_wallet.jinja",
                    wallet_data=wallet_importer.wallet_json,
                    wallet_type=wallet_importer.wallet_type,
                    wallet_name=wallet_importer.wallet_name,
                    cosigners=wallet_importer.cosigners,
                    unknown_cosigners=wallet_importer.unknown_cosigners,
                    unknown_cosigners_types=wallet_importer.unknown_cosigners_types,
                    sigs_required=wallet_importer.sigs_required,
                    sigs_total=wallet_importer.sigs_total,
                    specter=app.specter,
                    rand=rand,
                )
        if action == "device":
            cosigners = [
                app.specter.device_manager.get_by_alias(alias)
                for alias in request.form.getlist("devices")
            ]

            if not cosigners:
                return render_template(
                    "wallet/new_wallet/new_wallet.jinja",
                    wallet_type=wallet_type,
                    error=_(
                        "No device was selected. Please select a device to create the wallet for."
                    ),
                    specter=app.specter,
                    rand=rand,
                )
            devices = get_devices_with_keys_by_type(app, cosigners, wallet_type)
            for device in devices:
                if len(device.keys) == 0:
                    err = _(
                        "Device {} doesn't have keys matching this wallet type"
                    ).format(device.name)
                    break

            name = wallet_type.title()
            wallet_name = name
            i = 2
            while wallet_name in app.specter.wallet_manager.wallets_names:
                wallet_name = "%s %d" % (name, i)
                i += 1

            return render_template(
                "wallet/new_wallet/new_wallet_keys.jinja",
                cosigners=devices,
                wallet_type=wallet_type,
                sigs_total=len(devices),
                sigs_required=max(len(devices) * 2 // 3, 1),
                error=err,
                specter=app.specter,
                rand=rand,
            )
        if action == "key" and err is None:
            wallet_name = request.form["wallet_name"]
            address_type = request.form["type"]
            sigs_total = int(request.form.get("sigs_total", 1))
            sigs_required = int(request.form.get("sigs_required", 1))
            if wallet_name in app.specter.wallet_manager.wallets_names:
                err = _("Wallet already exists")
            if err:
                devices = [
                    app.specter.device_manager.get_by_alias(
                        request.form.get("cosigner{}".format(i))
                    )
                    for i in range(0, sigs_total)
                ]
                return render_template(
                    "wallet/new_wallet/new_wallet_keys.jinja",
                    cosigners=devices,
                    wallet_type=wallet_type,
                    sigs_total=len(devices),
                    sigs_required=max(len(devices) * 2 // 3, 1),
                    error=err,
                    specter=app.specter,
                    rand=rand,
                )

            keys = []
            cosigners = []
            devices = []
            for i in range(sigs_total):
                try:
                    key = request.form["key%d" % i]
                    cosigner_name = request.form["cosigner%d" % i]
                    cosigner = app.specter.device_manager.get_by_alias(cosigner_name)
                    cosigners.append(cosigner)
                    for k in cosigner.keys:
                        if k.original == key:
                            keys.append(k)
                            break
                except:
                    pass
            if len(keys) != sigs_total or len(cosigners) != sigs_total:
                err = _(
                    "No keys were selected for device, please try adding keys first"
                )
                devices = [
                    app.specter.device_manager.get_by_alias(
                        request.form.get("cosigner{}".format(i))
                    )
                    for i in range(0, sigs_total)
                ]
                return render_template(
                    "wallet/new_wallet/new_wallet_keys.jinja",
                    cosigners=devices,
                    wallet_type=wallet_type,
                    sigs_total=len(devices),
                    sigs_required=max(len(devices) * 2 // 3, 1),
                    error=err,
                    specter=app.specter,
                    rand=rand,
                )

            # create a wallet here
            try:
                wallet = app.specter.wallet_manager.create_wallet(
                    wallet_name, sigs_required, address_type, keys, cosigners
                )
            except Exception as e:
                handle_exception(e)
                err = _("Failed to create wallet. Error: {}").format(e)
                return render_template(
                    "wallet/new_wallet/new_wallet_keys.jinja",
                    cosigners=cosigners,
                    wallet_type=wallet_type,
                    sigs_total=len(devices),
                    sigs_required=max(len(devices) * 2 // 3, 1),
                    error=err,
                    specter=app.specter,
                    rand=rand,
                )

            app.logger.info("Created Wallet %s" % wallet_name)
            rescan_blockchain = "rescanblockchain" in request.form
            if rescan_blockchain:
                # old wallet - import more addresses
                wallet.keypoolrefill(0, wallet.IMPORT_KEYPOOL, change=False)
                wallet.keypoolrefill(0, wallet.IMPORT_KEYPOOL, change=True)
                if "utxo" in request.form.get("full_rescan_option"):
                    explorer = None
                    if "use_explorer" in request.form:
                        if request.form["explorer"] == "CUSTOM":
                            explorer = request.form["custom_explorer"]
                        else:
                            explorer = app.config["EXPLORERS_LIST"][
                                request.form["explorer"]
                            ]["url"]
                    wallet.rescanutxo(
                        explorer,
                        app.specter.requests_session(
                            explorer and explorer.endswith(".onion")
                        ),
                        app.specter.only_tor,
                    )
                    app.specter.info["utxorescan"] = 1
                    app.specter.utxorescanwallet = wallet.alias
                else:
                    app.logger.info("Rescanning Blockchain ...")
                    startblock = int(request.form["startblock"])
                    try:
                        wallet.rpc.rescanblockchain(startblock, no_wait=True)
                    except Exception as e:
                        handle_exception(e)
                        err = "%r" % e
                    wallet.getdata()
            return redirect(
                url_for("wallets_endpoint.receive", wallet_alias=wallet.alias)
                + "?newwallet=true"
            )
        if action == "preselected_device":
            return render_template(
                "wallet/new_wallet/new_wallet_keys.jinja",
                cosigners=[
                    app.specter.device_manager.get_by_alias(request.form["device"])
                ],
                wallet_type="simple",
                sigs_total=1,
                sigs_required=1,
                error=err,
                specter=app.specter,
                rand=rand,
            )

    return render_template(
        "wallet/new_wallet/new_wallet.jinja",
        wallet_type=wallet_type,
        error=err,
        specter=app.specter,
        rand=rand,
    )


################## Wallet pages #######################

###### Wallet index page ######
@wallets_endpoint.route("/wallet/<wallet_alias>/")
@login_required
def wallet(wallet_alias):
    wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
    if wallet.amount_total > 0:
        return redirect(url_for("wallets_endpoint.history", wallet_alias=wallet_alias))
    else:
        return redirect(url_for("wallets_endpoint.receive", wallet_alias=wallet_alias))


###### Wallet transaction history ######
@wallets_endpoint.route("/wallet/<wallet_alias>/history/", methods=["GET", "POST"])
@login_required
def history(wallet_alias):
    wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
    tx_list_type = "txlist"

    if request.method == "POST":
        action = request.form["action"]
        if action == "freezeutxo":
            wallet.toggle_freeze_utxo(request.form.getlist("selected_utxo"))
            tx_list_type = "utxo"
        elif action == "abandon_tx":
            txid = request.form["txid"]
            try:
                wallet.abandontransaction(txid)
            except SpecterError as e:
                flash(str(e), "error")

    # update balances in the wallet
    app.specter.check_blockheight()
    wallet.update_balance()
    wallet.check_utxo()

    return render_template(
        "wallet/history/wallet_history.jinja",
        wallet_alias=wallet_alias,
        wallet=wallet,
        tx_list_type=tx_list_type,
        specter=app.specter,
        rand=rand,
        services=app.specter.service_manager.services,
    )


###### Wallet receive ######


@wallets_endpoint.route("/wallet/<wallet_alias>/receive/", methods=["GET", "POST"])
@login_required
@check_wallet
def receive(wallet_alias):
    wallet: Wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
    if request.method == "POST":
        action = request.form["action"]
        if action == "newaddress":
            wallet.getnewaddress()
        elif action == "updatelabel":
            label = request.form["label"]
            wallet.setlabel(wallet.address, label)
    # check that current address is unused
    # and generate new one if it is
    wallet.check_unused()
    return render_template(
        "wallet/receive/wallet_receive.jinja",
        wallet_alias=wallet_alias,
        wallet=wallet,
        specter=app.specter,
        rand=rand,
    )


###### Wallet send ######


@wallets_endpoint.route("/wallet/<wallet_alias>/send")
@login_required
def send(wallet_alias):
    wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
    if len(wallet.pending_psbts) > 0:
        return redirect(
            url_for("wallets_endpoint.send_pending", wallet_alias=wallet_alias)
        )
    else:
        return redirect(url_for("wallets_endpoint.send_new", wallet_alias=wallet_alias))


@wallets_endpoint.route("/wallet/<wallet_alias>/send/new", methods=["GET", "POST"])
@login_required
def send_new(wallet_alias):
    wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
    # update balances in the wallet
    wallet.update_balance()
    # update utxo list for coin selection
    wallet.check_utxo()
    psbt = None
    addresses = [""]
    labels = [""]
    amounts = [0]
    amount_units = ["btc"]
    err = None
    ui_option = "ui"
    recipients_txt = ""
    fillform = False
    subtract = False
    subtract_from = 0
    fee_options = "dynamic"
    rbf = not app.specter.is_liquid
    rbf_utxo = []
    rbf_tx_id = ""
    selected_coins = request.form.getlist("coinselect")
    # Additional server side check not to use frozen UTXO as a precaution
    frozen_utxo = wallet.frozen_utxo
    for utxo in selected_coins:
        if utxo in frozen_utxo:
            selected_coins.remove(utxo)
            flash(f"You've selected a frozen UTXO for a transaction.", "error")
            return redirect(
                url_for("wallets_endpoint.history", wallet_alias=wallet_alias)
            )

    if request.method == "POST":
        action = request.form.get("action")
        rbf_tx_id = request.form.get("rbf_tx_id", "")
        if action == "createpsbt":
            psbt_creator = PsbtCreator(
                app.specter,
                wallet,
                request.form.get("ui_option", "ui"),
                request_form=request.form,
                recipients_txt=request.form["recipients"],
                recipients_amount_unit=request.form.get("amount_unit_text"),
            )
            psbt = psbt_creator.create_psbt(wallet)
            return render_template(
                "wallet/send/sign/wallet_send_sign_psbt.jinja",
                psbt=psbt,
                labels=labels,
                wallet_alias=wallet_alias,
                wallet=wallet,
                specter=app.specter,
                rand=rand,
            )

        elif action in ["rbf", "rbf_cancel"]:
            try:
                rbf_tx_id = request.form["rbf_tx_id"]
                rbf_fee_rate = float(request.form["rbf_fee_rate"])

                if action == "rbf":
                    psbt = wallet.bumpfee(rbf_tx_id, rbf_fee_rate)
                elif action == "rbf_cancel":
                    psbt = wallet.canceltx(rbf_tx_id, rbf_fee_rate)
                else:
                    raise SpecterError("Invalid action")

                if psbt["fee_rate"] - rbf_fee_rate > wallet.MIN_FEE_RATE / 10:
                    flash(
                        _(
                            "We had to increase the fee rate from {} to {} sat/vbyte"
                        ).format(rbf_fee_rate, psbt["fee_rate"])
                    )
                return render_template(
                    "wallet/send/sign/wallet_send_sign_psbt.jinja",
                    psbt=psbt,
                    labels=[],
                    wallet_alias=wallet_alias,
                    wallet=wallet,
                    specter=app.specter,
                    rand=rand,
                )
            except Exception as e:
                handle_exception(e)
                flash(_("Failed to perform RBF. Error: {}").format(e), "error")
                return redirect(
                    url_for("wallets_endpoint.history", wallet_alias=wallet_alias)
                )
        elif action == "rbf_edit":
            try:
                decoded_tx = wallet.decode_tx(rbf_tx_id)
                addresses = decoded_tx["addresses"]
                amounts = decoded_tx["amounts"]
                amount_units = decoded_tx.get("assets", ["btc"] * len(addresses))
                # get_label returns a label or address if no label is set
                labels = [wallet.getlabel(addr) for addr in addresses]
                # set empty label to addresses that don't have labels
                labels = [
                    label if label != addr else ""
                    for addr, label in zip(addresses, labels)
                ]

                selected_coins = [
                    f"{utxo['txid']}, {utxo['vout']}"
                    for utxo in decoded_tx["used_utxo"]
                ]
                fee_rate = float(request.form["rbf_fee_rate"])
                fee_options = "manual"
                rbf = True
                fillform = True
            except Exception as e:
                handle_exception(e)
                flash(_("Failed to perform RBF. Error: {}").format(e), "error")
        elif action == "signhotwallet":
            passphrase = request.form["passphrase"]
            psbt = json.loads(request.form["psbt"])
            current_psbt = wallet.pending_psbts[psbt["tx"]["txid"]]
            b64psbt = str(current_psbt)
            device = request.form["device"]
            if "devices_signed" not in psbt or device not in psbt["devices_signed"]:
                try:
                    # get device and sign with it
                    signed_psbt = app.specter.device_manager.get_by_alias(
                        device
                    ).sign_psbt(b64psbt, wallet, passphrase)
                    raw = None
                    if signed_psbt["complete"]:
                        raw = wallet.rpc.finalizepsbt(b64psbt)
                    current_psbt.update(signed_psbt["psbt"], raw)
                    signed_psbt = signed_psbt["psbt"]
                    psbt = current_psbt.to_dict()
                except Exception as e:
                    handle_exception(e)
                    signed_psbt = None
                    flash(_("Failed to sign PSBT: {}").format(e), "error")
            else:
                signed_psbt = None
                flash(_("Device already signed the PSBT"), "error")
            return render_template(
                "wallet/send/sign/wallet_send_sign_psbt.jinja",
                signed_psbt=signed_psbt,
                psbt=psbt,
                labels=labels,
                wallet_alias=wallet_alias,
                wallet=wallet,
                specter=app.specter,
                rand=rand,
            )
        elif action == "fillform":
            # TODO: Not yet used. Remove if the use case doesn't happen.
            # can be used to recommend a transaction from a service (goind to an exchange or so)
            addresses = request.form.getlist("addresses[]")
            labels = request.form.getlist("labels[]")
            amounts = request.form.getlist("amounts[]")
            fillform = True

    if rbf_tx_id:
        try:
            rbf_utxo = wallet.get_rbf_utxo(rbf_tx_id)
        except Exception as e:
            handle_exception(e)
            flash(_("Failed to get RBF coins. Error: {}").format(e), "error")

    show_advanced_settings = (
        ui_option != "ui" or subtract or fee_options != "dynamic" or not rbf
    )
    wallet_utxo = wallet.utxo
    if app.specter.is_liquid:
        for tx in wallet_utxo + rbf_utxo:
            if "asset" in tx:
                tx["assetlabel"] = app.specter.asset_label(tx.get("asset"))
    return render_template(
        "wallet/send/new/wallet_send.jinja",
        psbt=psbt,
        ui_option=ui_option,
        recipients_txt=recipients_txt,
        addresses=addresses,
        labels=labels,
        amounts=amounts,
        fillform=fillform,
        recipients=list(zip(addresses, amounts, amount_units, labels)),
        subtract=subtract,
        subtract_from=subtract_from,
        fee_options=fee_options,
        rbf=rbf,
        selected_coins=selected_coins,
        show_advanced_settings=show_advanced_settings,
        rbf_utxo=rbf_utxo,
        rbf_tx_id=rbf_tx_id,
        wallet_utxo=wallet_utxo,
        wallet_alias=wallet_alias,
        wallet=wallet,
        specter=app.specter,
        rand=rand,
        error=err,
    )


@wallets_endpoint.route("/wallet/<wallet_alias>/send/pending/", methods=["GET", "POST"])
@login_required
def send_pending(wallet_alias):
    wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
    if request.method == "POST":
        action = request.form["action"]
        if action == "deletepsbt":
            try:
                wallet.delete_pending_psbt(
                    json.loads(request.form["pending_psbt"])["tx"]["txid"]
                )
            except Exception as e:
                handle_exception(e)
                flash(_("Could not delete Pending PSBT!"), "error")
        elif action == "openpsbt":
            psbt = json.loads(request.form["pending_psbt"])
            return render_template(
                "wallet/send/sign/wallet_send_sign_psbt.jinja",
                psbt=psbt,
                wallet_alias=wallet_alias,
                wallet=wallet,
                specter=app.specter,
                rand=rand,
            )
    pending_psbts = wallet.pending_psbts_dict()
    return render_template(
        "wallet/send/pending/wallet_sendpending.jinja",
        pending_psbts=pending_psbts,
        wallet_alias=wallet_alias,
        wallet=wallet,
        specter=app.specter,
    )


@wallets_endpoint.route("/wallet/<wallet_alias>/send/import", methods=["GET", "POST"])
@login_required
def import_psbt(wallet_alias):
    wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
    if request.method == "POST":
        action = request.form["action"]
        if action == "importpsbt":
            try:
                b64psbt = "".join(request.form["rawpsbt"].split())
                psbt = wallet.importpsbt(
                    convert_rawtransaction_to_psbt(wallet.rpc, b64psbt)
                    if is_hex(b64psbt)
                    else b64psbt
                )
            except Exception as e:
                handle_exception(e)
                flash(_("Could not import PSBT: {}").format(e), "error")
                return redirect(
                    url_for("wallets_endpoint.import_psbt", wallet_alias=wallet_alias)
                )
            return render_template(
                "wallet/send/sign/wallet_send_sign_psbt.jinja",
                psbt=psbt,
                wallet_alias=wallet_alias,
                wallet=wallet,
                specter=app.specter,
                rand=rand,
            )
    err = None
    return render_template(
        "wallet/send/import/wallet_importpsbt.jinja",
        wallet_alias=wallet_alias,
        wallet=wallet,
        specter=app.specter,
        rand=rand,
        error=err,
    )


###### Wallet addresses ######


@wallets_endpoint.route("/wallet/<wallet_alias>/addresses/", methods=["GET"])
@login_required
def addresses(wallet_alias):
    """Show informations about cached addresses (wallet._addresses) of the <wallet_alias>.
    It updates balances in the wallet before renderization in order to show updated UTXO and
    balance of each address."""
    wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)

    # update balances in the wallet
    app.specter.check_blockheight()
    wallet.update_balance()
    wallet.check_utxo()

    return render_template(
        "wallet/addresses/wallet_addresses.jinja",
        wallet_alias=wallet_alias,
        wallet=wallet,
        specter=app.specter,
        rand=rand,
        services=app.specter.service_manager.services,
    )


###### Wallet settings ######


@wallets_endpoint.route("/wallet/<wallet_alias>/settings/", methods=["GET", "POST"])
@login_required
def settings(wallet_alias):
    wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
    error = None
    if request.method == "POST":
        action = request.form["action"]
        if action == "rescanblockchain":
            startblock = int(request.form["startblock"])
            try:
                delete_file(wallet._transactions.path)
                wallet.fetch_transactions()

                # This rpc call does not seem to return a result; use no_wait to ignore timeout errors
                wallet.rpc.rescanblockchain(startblock, no_wait=True)
            except Exception as e:
                handle_exception(e)
                error = "%r" % e
            wallet.getdata()
        elif action == "abortrescan":
            res = wallet.rpc.abortrescan()
            if not res:
                error = _("Failed to abort rescan. Maybe already complete?")
            wallet.getdata()
        elif action == "rescanutxo":
            explorer = None
            if "use_explorer" in request.form:
                if request.form["explorer"] == "CUSTOM":
                    explorer = request.form["custom_explorer"]
                else:
                    explorer = app.config["EXPLORERS_LIST"][request.form["explorer"]][
                        "url"
                    ]

            wallet.rescanutxo(
                explorer,
                app.specter.requests_session(explorer and explorer.endswith(".onion")),
                app.specter.only_tor,
            )
            app.specter.info["utxorescan"] = 1
            app.specter.utxorescanwallet = wallet.alias
            flash(
                "Rescan started. Check the status bar on the left for progress and/or the logs for potential issues."
            )
        elif action == "abortrescanutxo":
            app.specter.node.abortrescanutxo()
            app.specter.info["utxorescan"] = None
            app.specter.utxorescanwallet = None
            flash(_("Successfully aborted the UTXO rescan"))
        elif action == "import_address_labels":
            address_labels = request.form["address_labels_data"]
            imported_addresses_len = wallet.import_address_labels(address_labels)
            if imported_addresses_len > 1:
                flash(f"Successfully imported {imported_addresses_len} address labels.")
            elif imported_addresses_len == 1:
                flash(f"Successfully imported {imported_addresses_len} address label.")
            else:
                flash("No address labels were imported.")
        elif action == "keypoolrefill":
            delta = int(request.form["keypooladd"])
            wallet.keypoolrefill(wallet.keypool, wallet.keypool + delta)
            wallet.keypoolrefill(
                wallet.change_keypool, wallet.change_keypool + delta, change=True
            )
            wallet.getdata()
        elif action == "deletewallet":
            app.specter.wallet_manager.delete_wallet(
                wallet, app.specter.bitcoin_datadir, app.specter.chain
            )
            response = redirect(url_for("index"))
            return response
        elif action == "rename":
            wallet_name = request.form["newtitle"]
            if not wallet_name:
                flash(_("Wallet name cannot be empty"), "error")
            elif wallet_name == wallet.name:
                pass
            elif wallet_name in app.specter.wallet_manager.wallets_names:
                flash(_("Wallet already exists"), "error")
            else:
                app.specter.wallet_manager.rename_wallet(wallet, wallet_name)

    scroll_to_rescan_blockchain = request.args.get("rescan_blockchain")
    return render_template(
        "wallet/settings/wallet_settings.jinja",
        purposes=purposes,
        wallet_alias=wallet_alias,
        wallet=wallet,
        specter=app.specter,
        rand=rand,
        error=error,
        scroll_to_rescan_blockchain=scroll_to_rescan_blockchain,
    )



import traceback


class JSONRPC:
    """
    Base JSON-RPC class. Add methods to self.exposed_rpc
    to make it available with jsonrpc() call.
    """

    def __init__(self):
        self.exposed_rpc = {}

    def jsonrpc(self, request):
        """Processes json-rpc request"""
        # if it is a list (not bundled) run one by one
        if isinstance(request, list):
            return [self.jsonrpc(req) for req in request]
        response = {"jsonrpc": "2.0", "id": request["id"] if "id" in request else None}
        if "method" not in request:
            response["error"] = {
                "code": -32600,
                "message": "Invalid Request. Request must specify a 'method'.",
            }
            return response
        if request["method"] not in self.exposed_rpc:
            response["error"] = {"code": -32601, "message": "Method not found."}
            return response
        method = self.exposed_rpc[request["method"]]
        try:
            if "params" not in request:
                response["result"] = method()
            elif isinstance(request["params"], list):
                response["result"] = method(*request["params"])  # list -> *args
            else:
                response["result"] = method(**request["params"])  # dict -> **kwargs
        except Exception as e:
            traceback.print_exc()
            response["error"] = {"code": -32000, "message": f"Internal error: {e}."}
        return response
